import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class NMTM(nn.Module):
    '''
        Learning Multilingual Topics with Neural Variational Inference. NLPCC 2020.

        Xiaobao Wu, Chunping Li, Yan Zhu, Yishu Miao.
    '''
    def __init__(self, Map_en2cn, Map_cn2en, vocab_size_en, vocab_size_cn, num_topics=50, en_units=200, dropout=0., lam=0.8):
        super().__init__()

        self.num_topics = num_topics
        self.lam = lam

        # V_en x V_cn
        self.Map_en2cn = nn.Parameter(torch.as_tensor(Map_en2cn).float(), requires_grad=False)

        # V_cn x V_en
        self.Map_cn2en = nn.Parameter(torch.as_tensor(Map_cn2en).float(), requires_grad=False)

        self.a = 1 * np.ones((1, int(num_topics))).astype(np.float32)
        self.mu2 = nn.Parameter(torch.as_tensor((np.log(self.a).T - np.mean(np.log(self.a), 1)).T), requires_grad=False)
        self.var2 = nn.Parameter(torch.as_tensor((((1.0 / self.a) * (1 - (2.0 / num_topics))).T + (1.0 / (num_topics * num_topics)) * np.sum(1.0 / self.a, 1)).T), requires_grad=False)

        self.decoder_bn_en = nn.BatchNorm1d(vocab_size_en, eps=0.001, momentum=0.001, affine=True)
        self.decoder_bn_en.weight.requires_grad = False

        self.decoder_bn_cn = nn.BatchNorm1d(vocab_size_cn, eps=0.001, momentum=0.001, affine=True)
        self.decoder_bn_cn.weight.requires_grad = False

        self.fc11_en = nn.Linear(vocab_size_en, en_units)
        self.fc11_cn = nn.Linear(vocab_size_cn, en_units)
        self.fc12 = nn.Linear(en_units, en_units)
        self.fc21 = nn.Linear(en_units, num_topics)
        self.fc22 = nn.Linear(en_units, num_topics)

        self.fc1_drop = nn.Dropout(dropout)
        self.z_drop = nn.Dropout(dropout)

        self.mean_bn = nn.BatchNorm1d(num_topics, eps=0.001, momentum=0.001, affine=True)
        self.mean_bn.weight.requires_grad = False

        self.logvar_bn = nn.BatchNorm1d(num_topics, eps=0.001, momentum=0.001, affine=True)
        self.logvar_bn.weight.requires_grad = False

        self.phi_en = nn.Parameter(nn.init.xavier_uniform_(torch.empty((num_topics, vocab_size_en))))
        self.phi_cn = nn.Parameter(nn.init.xavier_uniform_(torch.empty((num_topics, vocab_size_cn))))

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mu)

    def encode(self, x, lang):
        e1 = F.softplus(getattr(self, f'fc11_{lang}')(x))

        e1 = F.softplus(self.fc12(e1))
        e1 = self.fc1_drop(e1)
        mu = self.mean_bn(self.fc21(e1))
        logvar = self.logvar_bn(self.fc22(e1))
        theta = self.reparameterize(mu, logvar)
        theta = F.softmax(theta, dim=1)
        theta = self.z_drop(theta)
        return theta, mu, logvar

    def get_theta(self, x, lang):
        theta, mu, logvar = self.encode(x, lang)

        if self.training:
            return theta, mu, logvar
        else:
            return mu

    def get_beta(self):
        beta_en = self.lam * torch.matmul(self.phi_cn, self.Map_cn2en) + (1 - self.lam) * self.phi_en
        beta_cn = self.lam * torch.matmul(self.phi_en, self.Map_en2cn) + (1 - self.lam) * self.phi_cn
        return beta_en, beta_cn

    def decode(self, theta, lang):
        d1 = F.softmax(getattr(self, f'decoder_bn_{lang}')(torch.matmul(theta, getattr(self, f'beta_{lang}'))), dim=1)
        return d1

    def forward(self, x_en, x_cn):
        self.beta_en, self.beta_cn = self.get_beta()

        theta_en, mu_en, logvar_en = self.get_theta(x_en, lang='en')
        theta_cn, mu_cn, logvar_cn = self.get_theta(x_cn, lang='cn')

        x_recon_en = self.decode(theta_en, lang='en')
        x_recon_cn = self.decode(theta_cn, lang='cn')

        loss_en = self.loss_function(x_recon_en, x_en, mu_en, logvar_en)
        loss_cn = self.loss_function(x_recon_cn, x_cn, mu_cn, logvar_cn)

        loss = loss_en + loss_cn

        rst_dict = {
            'loss': loss
        }

        return rst_dict

    def loss_function(self, recon_x, x, mu, logvar):
        var = logvar.exp()
        var_division = var / self.var2
        diff = mu - self.mu2
        diff_term = diff * diff / self.var2
        logvar_division = self.var2.log() - logvar
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(1) - self.num_topics)

        RECON = -(x * (recon_x + 1e-10).log()).sum(1)

        LOSS = (RECON + KLD).mean()
        return LOSS



import torch
import torch.nn as nn


class UWE(nn.Module):
    def __init__(self, ETC, num_times, temperature, weight_UWE, neg_topk):
        super().__init__()

        self.ETC = ETC
        self.weight_UWE = weight_UWE
        self.num_times = num_times
        self.temperature = temperature
        self.neg_topk = neg_topk

    def forward(self, time_wordcount, beta, topic_embeddings, word_embeddings):
        assert(self.num_times == time_wordcount.shape[0])

        topk_indices = self.get_topk_indices(beta)

        loss_UWE = 0.
        cnt_valid_times = 0.
        for t in range(self.num_times):
            neg_idx = torch.where(time_wordcount[t] == 0)[0]

            time_topk_indices = topk_indices[t]
            neg_idx = list(set(neg_idx.cpu().tolist()).intersection(set(time_topk_indices.cpu().tolist())))
            neg_idx = torch.tensor(neg_idx).long().to(time_wordcount.device)

            if len(neg_idx) == 0:
                continue

            time_neg_WE = word_embeddings[neg_idx]

            # topic_embeddings[t]: K x D
            # word_embeddings[neg_idx]: |V_{neg}| x D
            loss_UWE += self.ETC.compute_loss(topic_embeddings[t], time_neg_WE, temperature=self.temperature, all_neg=True)
            cnt_valid_times += 1

        if cnt_valid_times > 0:
            loss_UWE *= (self.weight_UWE / cnt_valid_times)

        return loss_UWE

    def get_topk_indices(self, beta):
        # topk_indices: T x K x neg_topk
        topk_indices = torch.topk(beta, k=self.neg_topk, dim=-1).indices
        topk_indices = torch.flatten(topk_indices, start_dim=1)
        return topk_indices



import numpy as np
from tqdm import tqdm
from collections import defaultdict

import torch
from torch.optim.lr_scheduler import StepLR
from topmost.utils import _utils
from topmost.utils.logger import Logger


logger = Logger("WARNING")


# transform tensor list to numpy list
def to_nparray(tensor_list):
    return np.asarray([item.detach().cpu().numpy() for item in tensor_list], dtype=object)


class HierarchicalTrainer:
    def __init__(self,
                 model,
                 dataset,
                 num_top_words=15,
                 epochs=200,
                 learning_rate=0.002,
                 batch_size=200,
                 lr_scheduler=None,
                 lr_step_size=125,
                 log_interval=5,
                 verbose=False
                ):

        self.model = model
        self.dataset = dataset
        self.num_top_words = num_top_words
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.lr_scheduler = lr_scheduler
        self.lr_step_size = lr_step_size
        self.log_interval = log_interval

        self.verbose = verbose
        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

    def make_optimizer(self):
        args_dict = {
            'params': self.model.parameters(),
            'lr': self.learning_rate,
        }

        optimizer = torch.optim.Adam(**args_dict)
        return optimizer

    def make_lr_scheduler(self, optimizer):
        lr_scheduler = StepLR(optimizer, step_size=self.lr_step_size, gamma=0.5, verbose=self.verbose)
        return lr_scheduler

    def train(self):
        optimizer = self.make_optimizer()

        if self.lr_scheduler:
            logger.info("using lr_scheduler")
            lr_scheduler = self.make_lr_scheduler(optimizer)

        data_size = len(self.dataset.train_dataloader.dataset)

        for epoch in tqdm(range(1, self.epochs + 1), leave=False):
            self.model.train()
            loss_rst_dict = defaultdict(float)

            for batch_data in self.dataset.train_dataloader:

                rst_dict = self.model(batch_data)
                batch_loss = rst_dict['loss']

                optimizer.zero_grad()
                batch_loss.backward()
                optimizer.step()

                for key in rst_dict:
                    loss_rst_dict[key] += rst_dict[key] * len(batch_data)

            if self.lr_scheduler:
                lr_scheduler.step()

            if epoch % self.log_interval == 0:
                output_log = f'Epoch: {epoch:03d}'
                for key in loss_rst_dict:
                    output_log += f' {key}: {loss_rst_dict[key] / data_size :.3f}'

                logger.info(output_log)

        top_words = self.get_top_words()
        train_theta = self.test(self.dataset.train_data)

        return top_words, train_theta

    def test(self, bow):
        data_size = bow.shape[0]

        num_topics_list = self.model.num_topics_list
        theta_list = [list() for _ in range(len(num_topics_list))]
        all_idx = torch.split(torch.arange(data_size), self.batch_size)

        with torch.no_grad():
            self.model.eval()
            for idx in all_idx:
                batch_input = bow[idx]
                batch_theta_list = self.model.get_theta(batch_input)

                for layer_id in range(len(num_topics_list)):
                    theta_list[layer_id].extend(batch_theta_list[layer_id].cpu().numpy().tolist())

        theta = np.empty(len(num_topics_list), object)
        theta[:] = [np.asarray(item) for item in theta_list]

        return theta

    def get_phi(self):
        phi = to_nparray(self.model.get_phi_list())
        return phi

    def get_beta(self):
        beta_list = to_nparray(self.model.get_beta())
        return beta_list

    def get_top_words(self, num_top_words=None, annotation=False):
        if num_top_words is None:
            num_top_words = self.num_top_words

        beta = self.get_beta()
        top_words_list = list()

        for layer in range(beta.shape[0]):
            if self.verbose:
                print(f"======= Layer: {layer} number of topics: {beta[layer].shape[0]} =======")
            top_words = _utils.get_top_words(beta[layer], self.dataset.vocab, num_top_words, self.verbose)

            if not annotation:
                top_words_list.append(top_words)
            else:
                top_words_list.extend([f'L-{layer}_K-{k} {item}' for k, item in enumerate(top_words)])

        return top_words_list

    def export_theta(self):
        train_theta = self.test(self.dataset.train_data)
        test_theta = self.test(self.dataset.test_data)

        return train_theta, test_theta



"""
https://radimrehurek.com/gensim/models/hdpmodel.html
"""

import gensim
from gensim.models import HdpModel
from topmost.utils import _utils
from topmost.utils.logger import Logger


logger = Logger("WARNING")


class HDPGensimTrainer:
    def __init__(self,
                 dataset,
                 num_top_words=15,
                 max_chunks=None,
                 max_time=None,
                 chunksize=256,
                 kappa=1.0,
                 tau=64.0,
                 K=15,
                 T=150,
                 alpha=1,
                 gamma=1,
                 eta=0.01,
                 scale=1.0,
                 var_converge=0.0001,
                 verbose=False
                ):

        self.dataset = dataset
        self.num_top_words = num_top_words
        self.vocab_size = dataset.vocab_size
        self.max_chunks = max_chunks
        self.max_time = max_time
        self.chunksize = chunksize
        self.kappa = kappa
        self.tau = tau
        self.K = K
        self.T = T
        self.alpha = alpha
        self.gamma = gamma
        self.eta = eta
        self.scale = scale
        self.var_converge = var_converge

        self.verbose = verbose

    def train(self):
        train_bow = self.dataset.train_bow.astype("int32")
        id2word = dict(zip(range(self.vocab_size), self.dataset.vocab))
        corpus = gensim.matutils.Dense2Corpus(train_bow, documents_columns=False)

        self.model = HdpModel(
            corpus=corpus,
            id2word=id2word,
            max_chunks=self.max_chunks,
            max_time=self.max_time,
            chunksize=self.chunksize,
            kappa=self.kappa,
            tau=self.tau,
            K=self.K,
            T=self.T,
            alpha=self.alpha,
            gamma=self.gamma,
            eta=self.eta,
            scale=self.scale,
            var_converge=self.var_converge
        )

        top_words = self.get_top_words()
        train_theta = self.test(self.dataset.train_bow)

        return top_words, train_theta

    def test(self, bow):
        theta = list()
        beta = self.get_beta()
        corpus = gensim.matutils.Dense2Corpus(bow.astype('int32'), documents_columns=False)
        for doc in corpus:
            theta.append(self.model[doc])
        theta = gensim.matutils.corpus2dense(theta, num_docs=bow.shape[0], num_terms=beta.shape[0])
        theta = theta.transpose()
        return theta

    def get_beta(self):
        return self.model.get_topics()

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        beta = self.get_beta()
        top_words = _utils.get_top_words(beta, self.dataset.vocab, num_top_words, self.verbose)
        return top_words

    def export_theta(self):
        train_theta = self.test(self.dataset.train_bow)
        test_theta = self.test(self.dataset.test_bow)
        return train_theta, test_theta



from .basic.basic_trainer import BasicTrainer
from .basic.BERTopic_trainer import BERTopicTrainer
from .basic.FASTopic_trainer import FASTopicTrainer
from .basic.LDA_trainer import LDAGensimTrainer
from .basic.LDA_trainer import LDASklearnTrainer
from .basic.NMF_trainer import NMFGensimTrainer
from .basic.NMF_trainer import NMFSklearnTrainer

from .crosslingual.crosslingual_trainer import CrosslingualTrainer
from .dynamic.dynamic_trainer import DynamicTrainer

from .dynamic.DTM_trainer import DTMTrainer

from .hierarchical.hierarchical_trainer import HierarchicalTrainer
from .hierarchical.HDP_trainer import HDPGensimTrainer



import gensim
from gensim.models import nmf
from topmost.utils import _utils


class NMFGensimTrainer:
    def __init__(self,
                 dataset,
                 num_topics=50,
                 num_top_words=15,
                 max_iter=1
                ):
        self.dataset = dataset
        self.num_topics = num_topics
        self.num_top_words = num_top_words
        self.vocab_size = dataset.vocab_size
        self.max_iter = max_iter

    def train(self):
        train_bow = self.dataset.train_bow.astype("int32")
        id2word = dict(zip(range(self.vocab_size), self.dataset.vocab))
        corpus = gensim.matutils.Dense2Corpus(train_bow, documents_columns=False)

        self.model = nmf.Nmf(
            corpus=corpus,
            num_topics=self.num_topics,
            id2word=id2word,
            passes=self.max_iter
        )

        top_words = self.get_top_words()
        train_theta = self.test(self.dataset.train_bow)

        return top_words, train_theta

    def test(self, bow):
        corpus = gensim.matutils.Dense2Corpus(bow.astype('int64'), documents_columns=False)
        theta = gensim.matutils.corpus2dense(self.model.get_document_topics(corpus), num_docs=bow.shape[0], num_terms=self.num_topics)
        theta = theta.transpose()
        return theta

    def get_beta(self):
        return self.model.get_topics()

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        beta = self.get_beta()
        top_words = _utils.get_top_words(beta, self.dataset.vocab, num_top_words)
        return top_words

    def export_theta(self):
        train_theta = self.test(self.dataset.train_bow)
        test_theta = self.test(self.dataset.test_bow)
        return train_theta, test_theta


class NMFSklearnTrainer:
    def __init__(self,
                 model,
                 dataset,
                 num_top_words=15):
        self.model = model
        self.dataset = dataset
        self.num_top_words = num_top_words

    def train(self):
        train_bow = self.dataset.train_bow.astype('int64')
        self.model.fit(train_bow)

        top_words = self.get_top_words()
        train_theta = self.test(self.dataset.train_bow)
        return top_words, train_theta

    def test(self, bow):
        return self.model.transform(bow.astype('int64'))

    def get_beta(self):
        return self.model.components_

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        beta = self.get_beta()
        top_words = _utils.get_top_words(beta, self.dataset.vocab, num_top_words)
        return top_words

    def export_theta(self):
        train_theta = self.test(self.dataset.train_bow)
        test_theta = self.test(self.dataset.test_bow)
        return train_theta, test_theta



import numpy as np
from tqdm import tqdm
from collections import Counter, defaultdict
from sklearn.feature_extraction.text import CountVectorizer

from .build_hierarchy import build_hierarchy


def parse_item_info(topic_str):
    item_info = topic_str.split()[0]
    layer_id, topic_idx = (int(item.split('-')[1]) for item in item_info.split('_'))
    return layer_id, topic_idx


def convert_topicStr_to_dict(topic_str_list):
    """

    Args:
        topic_str_list: [L-0_K-0 w1 w2 w3 ..., L-0_K-1 w1 w2 w3 ...]. L indicates the layer, and K indicates the topic.
        keep_info (bool, optional): if keep the item info L-0_K-0. Defaults to False.

    Returns:
        hierarchical_topic_dict: {0: ["w1 w2...", "w1 w2..."], 1: ["w1 w2...", "w1 w2..."]}

    """

    hierarchical_topic_dict = defaultdict(list)

    for topic_str in topic_str_list:
        topic_str_items = topic_str.split()
        layer_id, k = parse_item_info(topic_str)

        dict_item = ' '.join(topic_str_items)

        hierarchical_topic_dict[layer_id].append(dict_item)

    return hierarchical_topic_dict


def compute_TD(texts):
    K = len(texts)
    T = len(texts[0].split())
    vectorizer = CountVectorizer()
    counter = vectorizer.fit_transform(texts).toarray()

    TF = counter.sum(axis=0)
    TD = (TF == 1).sum() / (K * T)

    return TD


def compute_CLNPMI(parent_diff_words, child_diff_words, all_bow, vocab):
    npmi_list = list()

    for p_w in parent_diff_words:
        flag_n = all_bow[:, vocab.index(p_w)] > 0
        p_n = np.sum(flag_n) / len(all_bow)

        for c_w in child_diff_words:
            flag_l = all_bow[:, vocab.index(c_w)] > 0
            p_l = np.sum(flag_l)
            p_nl = np.sum(flag_n * flag_l)

            if p_nl == len(all_bow):
                npmi_score = 1
            else:
                p_l = p_l / len(all_bow)
                p_nl = p_nl / len(all_bow)
                p_nl += 1e-10
                npmi_score = np.log(p_nl / (p_l * p_n)) / -np.log(p_nl)

            npmi_list.append(npmi_score)

    return npmi_list


def get_CLNPMI(PC_pair_groups, all_bow, vocab):
    CNPMI_list = list()
    for group in tqdm(PC_pair_groups):
        layer_CNPMI = list()
        for parent_topic, child_topic in group:
            parent_words = set(parent_topic.split())
            child_words = set(child_topic.split())

            inter = parent_words.intersection(child_words)
            parent_diff_words = list(parent_words.difference(inter))
            child_diff_words = list(child_words.difference(inter))

            npmi_list = compute_CLNPMI(parent_diff_words, child_diff_words, all_bow, vocab)

            # NOTE: assign -1 to the NPMI of repetitive word pairs
            num_repetition = (len(parent_words) - len(parent_diff_words)) * (len(child_words) - len(child_diff_words))
            npmi_list.extend([-1] * num_repetition)

            layer_CNPMI.extend(npmi_list)

        CNPMI_list.append(np.mean(layer_CNPMI))

    return CNPMI_list


def compute_diff_topic_pair(topic_str_a, topic_str_b):
    word_counter = Counter()
    topic_words_a = topic_str_a.split()
    topic_words_b = topic_str_b.split()
    word_counter.update(topic_words_a)
    word_counter.update(topic_words_b)
    diff = (np.asarray(list(word_counter.values())) == 1).sum() / (len(topic_words_a) + len(topic_words_b))
    return diff


def get_topics_difference(topic_pair_groups):
    diff_list = list()
    for groups in topic_pair_groups:
        layer_diff = list()
        for topic_a, topic_b in groups:
            diff = compute_diff_topic_pair(topic_a, topic_b)
            layer_diff.append(diff)
        diff_list.append(np.mean(layer_diff))

    return diff_list


# Given a list of child topics, find the nonchild topics based on their item info.
def extract_nonchild_topic_list(hierarchical_topic_dict, child_topic_list, num_topics_list):
    child_topic_idx_list = [parse_item_info(child_topic)[1] for child_topic in child_topic_list]
    layer_id, _ = parse_item_info(child_topic_list[0])
    # num_topic = beta_list[layer_id].shape[0]
    num_topic = num_topics_list[layer_id]
    nonchild_topic_idx_list = list(set(range(num_topic)) - set(child_topic_idx_list))
    nonchild_topic_list = np.asarray(hierarchical_topic_dict[layer_id])[nonchild_topic_idx_list].tolist()

    return nonchild_topic_list


def get_topic_pairs(topic_pairs, topic_hierarchy, hierarchical_topic_dict, num_topics_list, _type, layer_id=0):
    for parent_topic in topic_hierarchy.keys():
        if isinstance(topic_hierarchy[parent_topic], list):
            child_topic_list = topic_hierarchy[parent_topic]
        else:
            child_topic_list = list(topic_hierarchy[parent_topic].keys())

        if _type == 'PC':
            for child_topic in child_topic_list:
                topic_pairs[layer_id].append((parent_topic, child_topic))

        elif _type == 'PnonC':
            nonchild_topic_list = extract_nonchild_topic_list(hierarchical_topic_dict, child_topic_list, num_topics_list)
            for nonchild_topic in nonchild_topic_list:
                topic_pairs[layer_id].append((parent_topic, nonchild_topic))

        # Move to the next layer if more.
        if not isinstance(topic_hierarchy[parent_topic], list):
            get_topic_pairs(topic_pairs, topic_hierarchy[parent_topic], hierarchical_topic_dict, num_topics_list, _type, layer_id + 1)


# sibling_groups: length == num_layers
# each element in the list is a group of sibling topics at a layer.
def get_sibling_groups(topic_hierarchy, sibling_groups, layer_id=0):
    if isinstance(topic_hierarchy, list):
        sibling_groups[layer_id].append(topic_hierarchy)
    else:
        # sibling topics at this layer
        sibling_groups[layer_id].append(list(topic_hierarchy.keys()))
        # sibling topics at next layer
        for parent_topic in topic_hierarchy.keys():
            get_sibling_groups(topic_hierarchy[parent_topic], sibling_groups, layer_id + 1)


def get_Sibling_TD(sibling_groups):
    sibling_TD = list()
    for group in sibling_groups:
        layer_sibling_TD = list()
        for sibling_topics in group:
            TD = compute_TD(sibling_topics)
            layer_sibling_TD.append(TD)
        sibling_TD.append(np.mean(layer_sibling_TD))
    return sibling_TD


def get_Sibling_NPMI(sibling_groups, all_bow, vocab):
    sibling_NPMI = list()
    for group in sibling_groups:
        layer_pairs = list()
        for sibling_topics in group:
            sibling_num = len(sibling_topics)
            for i in range(sibling_num):
                for j in range(i + 1, sibling_num):
                    layer_pairs.append([sibling_topics[i], sibling_topics[j]])

        npmi = get_CLNPMI(layer_pairs, all_bow, vocab)
        sibling_NPMI.append(np.mean(npmi))
    return sibling_NPMI


def get_topic_groups(hierarchical_topic_dict, topic_hierarchy, beta_list):
    num_layers = len(beta_list)
    num_topics_list = [item.shape[0] for item in beta_list]

    PC_pair_groups = [list() for _ in range(num_layers - 1)]
    PnonC_pair_groups = [list() for _ in range(num_layers - 1)]

    get_topic_pairs(PC_pair_groups, topic_hierarchy, hierarchical_topic_dict, num_topics_list, _type='PC')
    get_topic_pairs(PnonC_pair_groups, topic_hierarchy, hierarchical_topic_dict, num_topics_list, _type='PnonC')

    sibling_groups = [list() for _ in range(num_layers)]
    get_sibling_groups(topic_hierarchy, sibling_groups)

    # Because hierarhcial_topic_dict contains item info of each topic (Layer-0_K-20)
    # Remove these item info from topic strings
    PC_pair_groups = clean_group_info(PC_pair_groups)
    PnonC_pair_groups = clean_group_info(PnonC_pair_groups)
    sibling_groups = clean_group_info(sibling_groups)
    return PC_pair_groups, PnonC_pair_groups, sibling_groups


def clean_group_info(groups):
    clean_groups = list()
    for layer_group in groups:
        layer_clean_group = list()
        for topic_str_list in layer_group:
            layer_clean_group.append(clean_info(topic_str_list))
        clean_groups.append(layer_clean_group)
    return clean_groups


# remove item info of topic strings
# L-0_K-0 w1 w2 ===> w1 w2.
def clean_info(topic_str_list):
    clean_list = [" ".join(item.split()[1:]) for item in topic_str_list]
    return clean_list


def hierarchy_quality(vocab, reference_bow, topic_str_list, beta_list, phi_list):
    hierarchical_topic_dict = convert_topicStr_to_dict(topic_str_list)
    topic_hierarchy = build_hierarchy(hierarchical_topic_dict, phi_list)

    PC_pair_groups, PnonC_pair_groups, sibling_groups = get_topic_groups(hierarchical_topic_dict, topic_hierarchy, beta_list)

    # Parent and Child topic Coherence (PCC)
    CLNPMI = get_CLNPMI(PC_pair_groups, reference_bow, vocab)

    # Parent and Child topic Diversity (PCD)
    PC_TD = get_topics_difference(PC_pair_groups)

    # Sibling Topic Diversity (SD)
    Sibling_TD = get_Sibling_TD(sibling_groups)

    # Parent and non-Child Topic Diversity (PnCD)
    PnonC_TD = get_topics_difference(PnonC_pair_groups)

    rst = {
        'PCC': np.mean(CLNPMI),
        'PCD': np.mean(PC_TD),
        'Sibling_TD': np.mean(Sibling_TD),
        'PnCD': np.mean(PnonC_TD)
    }

    return rst, topic_hierarchy



import torch
from torch.utils.data import Dataset, DataLoader
import numpy as np
import scipy.sparse
import scipy.io
from . import file_utils


class _SequentialDataset(Dataset):
    def __init__(self, bow, times, time_wordfreq):
        super().__init__()
        self.bow = bow
        self.times = times
        self.time_wordfreq = time_wordfreq

    def __len__(self):
        return len(self.bow)

    def __getitem__(self, index):
        return_dict = {
            'bow': self.bow[index],
            'times': self.times[index],
            'time_wordfreq': self.time_wordfreq[self.times[index]],
        }

        return return_dict


class DynamicDataset:
    def __init__(self, dataset_dir, batch_size=200, read_labels=False, device='cpu', as_tensor=True):

        self.load_data(dataset_dir, read_labels)

        self.vocab_size = len(self.vocab)
        self.train_size = len(self.train_bow)
        self.num_times = len(np.unique(self.train_times))
        self.train_time_wordfreq = self.get_time_wordfreq(self.train_bow, self.train_times)

        print('train size: ', len(self.train_bow))
        print('test size: ', len(self.test_bow))
        print('vocab size: ', len(self.vocab))
        print('average length: {:.3f}'.format(self.train_bow.sum(1).mean().item()))
        print('num of each time slice: ', self.num_times, np.bincount(self.train_times))

        if as_tensor:
            self.train_bow = torch.from_numpy(self.train_bow).float().to(device)
            self.test_bow = torch.from_numpy(self.test_bow).float().to(device)
            self.train_times = torch.from_numpy(self.train_times).long().to(device)
            self.test_times = torch.from_numpy(self.test_times).long().to(device)
            self.train_time_wordfreq = torch.from_numpy(self.train_time_wordfreq).float().to(device)

            self.train_dataset = _SequentialDataset(self.train_bow, self.train_times, self.train_time_wordfreq)
            self.test_dataset = _SequentialDataset(self.test_bow, self.test_times, self.train_time_wordfreq)

            self.train_dataloader = DataLoader(self.train_dataset, batch_size=batch_size, shuffle=True)

    def load_data(self, path, read_labels):
        self.train_bow = scipy.sparse.load_npz(f'{path}/train_bow.npz').toarray().astype('float32')
        self.test_bow = scipy.sparse.load_npz(f'{path}/test_bow.npz').toarray().astype('float32')
        self.word_embeddings = scipy.sparse.load_npz(f'{path}/word_embeddings.npz').toarray().astype('float32')

        self.train_texts = file_utils.read_text(f'{path}/train_texts.txt')
        self.test_texts = file_utils.read_text(f'{path}/test_texts.txt')

        self.train_times = np.loadtxt(f'{path}/train_times.txt').astype('int32')
        self.test_times = np.loadtxt(f'{path}/test_times.txt').astype('int32')

        self.vocab = file_utils.read_text(f'{path}/vocab.txt')

        self.pretrained_WE = scipy.sparse.load_npz(f'{path}/word_embeddings.npz').toarray().astype('float32')

        if read_labels:
            self.train_labels = np.loadtxt(f'{path}/train_labels.txt').astype('int32')
            self.test_labels = np.loadtxt(f'{path}/test_labels.txt').astype('int32')

    # word frequency at each time slice.
    def get_time_wordfreq(self, bow, times):
        train_time_wordfreq = np.zeros((self.num_times, self.vocab_size))
        for time in range(self.num_times):
            idx = np.where(times == time)[0]
            train_time_wordfreq[time] += bow[idx].sum(0)
        cnt_times = np.bincount(times)

        train_time_wordfreq = train_time_wordfreq / cnt_times[:, np.newaxis]
        return train_time_wordfreq




import torch
import torch.nn as nn
import torch.nn.functional as F


class TopicDistQuant(nn.Module):
    '''
        Short Text Topic Modeling with Topic Distribution Quantization and Negative Sampling Decoder. EMNLP 2020

        Xiaobao Wu, Chunping Li, Yan Zhu, Yishu Miao
    '''
    def __init__(self, num_embeddings, embedding_dim, commitment_cost=0.1):
        super().__init__()

        self._embedding_dim = embedding_dim
        self._num_embeddings = num_embeddings
        self._embedding = nn.Embedding(self._num_embeddings, self._embedding_dim)
        self._embedding.weight.data.copy_(torch.eye(embedding_dim))
        self._commitment_cost = commitment_cost

    def forward(self, inputs):
        # Calculate distances
        # NOTE: Do not use torch.cdist. It has unknown bugs.
        distances = (torch.sum(inputs**2, dim=1, keepdim=True) 
                   + torch.sum(self._embedding.weight**2, dim=1)
                   - 2 * torch.matmul(inputs, self._embedding.weight.t()))

        # Encoding
        encoding_indices = torch.argmin(distances, dim=1)

        # Quantize and unflatten
        quantized = self._embedding(encoding_indices)

        # Loss
        e_latent_loss = F.mse_loss(quantized.detach(), inputs, reduction='none').sum(axis=1).mean()
        q_latent_loss = F.mse_loss(quantized, inputs.detach(), reduction='none').sum(axis=1).mean()
        loss = q_latent_loss + self._commitment_cost * e_latent_loss

        quantized = inputs + (quantized - inputs).detach()

        rst = {
            'loss': loss,
            'quantized': quantized,
            'encoding_indices': encoding_indices,
        }

        return rst



import torch.nn as nn


def _get_activation_fn(activation):
    if activation == "relu":
        return nn.ReLU()
    elif activation == "softplus":
        return nn.Softplus()
    elif activation == "tanh":
        return nn.Tanh()
    else:
        raise RuntimeError("activation should be relu/tanh/softplus, not {}".format(activation))


class ResBlock(nn.Module):
    """Simple MLP block with residual connection.

    Args:
        in_features: the feature dimension of each output sample.
        out_features: the feature dimension of each output sample.
        activation: the activation function of intermediate layer, relu or gelu.
    """

    def __init__(self, in_features, out_features, activation="relu"):
        super(ResBlock, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.fc1 = nn.Linear(in_features, out_features)
        self.fc2 = nn.Linear(out_features, out_features)

        self.bn = nn.BatchNorm1d(out_features)
        self.activation = _get_activation_fn(activation)

    def forward(self, x):
        if self.in_features == self.out_features:
            out = self.fc2(self.activation(self.fc1(x)))
            return self.activation(self.bn(x + out))
        else:
            x = self.fc1(x)
            out = self.fc2(self.activation(x))
            return self.activation(self.bn(x + out))



import os
import json


def make_dir(path):
    os.makedirs(path, exist_ok=True)


def read_text(path):
    texts = list()
    with open(path, 'r', encoding='utf-8', errors='ignore') as file:
        for line in file:
            texts.append(line.strip())
    return texts


def save_text(texts, path):
    with open(path, 'w', encoding='utf-8') as file:
        for text in texts:
            file.write(text.strip() + '\n')


def read_jsonlist(path):
    data = list()
    with open(path, 'r', encoding='utf-8') as input_file:
        for line in input_file:
            data.append(json.loads(line))
    return data


def save_jsonlist(list_of_json_objects, path, sort_keys=True):
    with open(path, 'w', encoding='utf-8') as output_file:
        for obj in list_of_json_objects:
            output_file.write(json.dumps(obj, sort_keys=sort_keys) + '\n')


def split_text_word(texts):
    texts = [text.split() for text in texts]
    return texts



import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from .ECR import ECR


class ECRTM(nn.Module):
    '''
        Effective Neural Topic Modeling with Embedding Clustering Regularization. ICML 2023

        Xiaobao Wu, Xinshuai Dong, Thong Thanh Nguyen, Anh Tuan Luu.
    '''
    def __init__(self, vocab_size, num_topics=50, en_units=200, dropout=0., pretrained_WE=None, embed_size=200, beta_temp=0.2, weight_loss_ECR=100.0, sinkhorn_alpha=20.0, sinkhorn_max_iter=1000):
        super().__init__()

        self.num_topics = num_topics
        self.beta_temp = beta_temp

        self.a = 1 * np.ones((1, num_topics)).astype(np.float32)
        self.mu2 = nn.Parameter(torch.as_tensor((np.log(self.a).T - np.mean(np.log(self.a), 1)).T))
        self.var2 = nn.Parameter(torch.as_tensor((((1.0 / self.a) * (1 - (2.0 / num_topics))).T + (1.0 / (num_topics * num_topics)) * np.sum(1.0 / self.a, 1)).T))

        self.mu2.requires_grad = False
        self.var2.requires_grad = False

        self.fc11 = nn.Linear(vocab_size, en_units)
        self.fc12 = nn.Linear(en_units, en_units)
        self.fc21 = nn.Linear(en_units, num_topics)
        self.fc22 = nn.Linear(en_units, num_topics)
        self.fc1_dropout = nn.Dropout(dropout)
        self.theta_dropout = nn.Dropout(dropout)

        self.mean_bn = nn.BatchNorm1d(num_topics)
        self.mean_bn.weight.requires_grad = False
        self.logvar_bn = nn.BatchNorm1d(num_topics)
        self.logvar_bn.weight.requires_grad = False
        self.decoder_bn = nn.BatchNorm1d(vocab_size, affine=True)
        self.decoder_bn.weight.requires_grad = False

        if pretrained_WE is not None:
            self.word_embeddings = torch.from_numpy(pretrained_WE).float()
        else:
            self.word_embeddings = nn.init.trunc_normal_(torch.empty(vocab_size, embed_size))
        self.word_embeddings = nn.Parameter(F.normalize(self.word_embeddings))

        self.topic_embeddings = torch.empty((num_topics, self.word_embeddings.shape[1]))
        nn.init.trunc_normal_(self.topic_embeddings, std=0.1)
        self.topic_embeddings = nn.Parameter(F.normalize(self.topic_embeddings))

        self.ECR = ECR(weight_loss_ECR, sinkhorn_alpha, sinkhorn_max_iter)

    def get_beta(self):
        dist = self.pairwise_euclidean_distance(self.topic_embeddings, self.word_embeddings)
        beta = F.softmax(-dist / self.beta_temp, dim=0)
        return beta

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return mu + (eps * std)
        else:
            return mu

    def encode(self, input):
        e1 = F.softplus(self.fc11(input))
        e1 = F.softplus(self.fc12(e1))
        e1 = self.fc1_dropout(e1)
        mu = self.mean_bn(self.fc21(e1))
        logvar = self.logvar_bn(self.fc22(e1))
        z = self.reparameterize(mu, logvar)
        theta = F.softmax(z, dim=1)

        loss_KL = self.compute_loss_KL(mu, logvar)

        return theta, loss_KL

    def get_theta(self, input):
        theta, loss_KL = self.encode(input)
        if self.training:
            return theta, loss_KL
        else:
            return theta

    def compute_loss_KL(self, mu, logvar):
        var = logvar.exp()
        var_division = var / self.var2
        diff = mu - self.mu2
        diff_term = diff * diff / self.var2
        logvar_division = self.var2.log() - logvar
        # KLD: N*K
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(axis=1) - self.num_topics)
        KLD = KLD.mean()
        return KLD

    def get_loss_ECR(self):
        cost = self.pairwise_euclidean_distance(self.topic_embeddings, self.word_embeddings)
        loss_ECR = self.ECR(cost)
        return loss_ECR

    def pairwise_euclidean_distance(self, x, y):
        cost = torch.sum(x ** 2, axis=1, keepdim=True) + torch.sum(y ** 2, dim=1) - 2 * torch.matmul(x, y.t())
        return cost

    def forward(self, input):
        theta, loss_KL = self.encode(input)
        beta = self.get_beta()

        recon = F.softmax(self.decoder_bn(torch.matmul(theta, beta)), dim=-1)
        recon_loss = -(input * recon.log()).sum(axis=1).mean()

        loss_TM = recon_loss + loss_KL

        loss_ECR = self.get_loss_ECR()
        loss = loss_TM + loss_ECR

        rst_dict = {
            'loss': loss,
            'loss_TM': loss_TM,
            'loss_ECR': loss_ECR
        }

        return rst_dict



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from topmost.models.Encoder import MLPEncoder
from .TAMI import TAMI


class InfoCTM(nn.Module):
    '''
        InfoCTM: A Mutual Information Maximization Perspective of Cross-lingual Topic Modeling. AAAI 2023

        Xiaobao Wu, Xinshuai Dong, Thong Nguyen, Chaoqun Liu, Liangming Pan, Anh Tuan Luu
    '''
    def __init__(self, trans_e2c, pretrain_word_embeddings_en, pretrain_word_embeddings_cn, vocab_size_en, vocab_size_cn, num_topics=50, en_units=200, dropout=0., temperature=0.2, pos_threshold=0.4, weight_MI=30.0):
        super().__init__()

        self.num_topics = num_topics

        self.encoder_en = MLPEncoder(vocab_size_en, num_topics, en_units, dropout)
        self.encoder_cn = MLPEncoder(vocab_size_cn, num_topics, en_units, dropout)

        self.a = 1 * np.ones((1, int(num_topics))).astype(np.float32)
        self.mu2 = nn.Parameter(torch.as_tensor((np.log(self.a).T - np.mean(np.log(self.a), 1)).T), requires_grad=False)
        self.var2 = nn.Parameter(torch.as_tensor((((1.0 / self.a) * (1 - (2.0 / num_topics))).T + (1.0 / (num_topics * num_topics)) * np.sum(1.0 / self.a, 1)).T), requires_grad=False)

        self.decoder_bn_en = nn.BatchNorm1d(vocab_size_en, affine=True)
        self.decoder_bn_en.weight.requires_grad = False
        self.decoder_bn_cn = nn.BatchNorm1d(vocab_size_cn, affine=True)
        self.decoder_bn_cn.weight.requires_grad = False

        self.phi_en = nn.Parameter(nn.init.xavier_uniform_(torch.empty((num_topics, vocab_size_en))))
        self.phi_cn = nn.Parameter(nn.init.xavier_uniform_(torch.empty((num_topics, vocab_size_cn))))

        self.TAMI = TAMI(temperature, weight_MI, pos_threshold, trans_e2c, pretrain_word_embeddings_en, pretrain_word_embeddings_cn)

    def get_beta(self):
        beta_en = self.phi_en
        beta_cn = self.phi_cn
        return beta_en, beta_cn

    def get_theta(self, x, lang):
        theta, mu, logvar = getattr(self, f'encoder_{lang}')(x)

        if self.training:
            return theta, mu, logvar
        else:
            return mu

    def decode(self, theta, beta, lang):
        bn = getattr(self, f'decoder_bn_{lang}')
        d1 = F.softmax(bn(torch.matmul(theta, beta)), dim=1)
        return d1

    def forward(self, x_en, x_cn):
        theta_en, mu_en, logvar_en = self.get_theta(x_en, lang='en')
        theta_cn, mu_cn, logvar_cn = self.get_theta(x_cn, lang='cn')

        beta_en, beta_cn = self.get_beta()

        loss = 0.

        x_recon_en = self.decode(theta_en, beta_en, lang='en')
        x_recon_cn = self.decode(theta_cn, beta_cn, lang='cn')
        loss_en = self.compute_loss_TM(x_recon_en, x_en, mu_en, logvar_en)
        loss_cn = self.compute_loss_TM(x_recon_cn, x_cn, mu_cn, logvar_cn)

        loss = loss_en + loss_cn

        fea_en = beta_en.T
        fea_cn = beta_cn.T
        loss_TAMI = self.TAMI(fea_en, fea_cn)

        loss += loss_TAMI

        rst_dict = {
            'loss': loss,
            'loss_TAMI': loss_TAMI
        }

        return rst_dict

    def compute_loss_TM(self, recon_x, x, mu, logvar):
        var = logvar.exp()
        var_division = var / self.var2
        diff = mu - self.mu2
        diff_term = diff * diff / self.var2
        logvar_division = self.var2.log() - logvar
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(1) - self.num_topics)

        RECON = -(x * (recon_x + 1e-10).log()).sum(1)

        LOSS = (RECON + KLD).mean()
        return LOSS



import pytest
import numpy as np

import sys
sys.path.append('../')

from topmost.data import download_20ng, file_utils
from topmost.preprocess import Preprocess


@pytest.fixture
def cache_path():
    return './pytest_cache/'


def test_datasets(cache_path):
    path = f'{cache_path}/20NG'
    download_20ng.download_save(output_dir=path)
    preprocess = Preprocess(vocab_size=5000, min_term=1, stopwords='snowball')
    rst = preprocess.preprocess_jsonlist(dataset_dir=path, label_name="group")
    preprocess.save(path, **rst)



import pytest
import sys

sys.path.append("../")

import torch
import topmost
from topmost import eva, Preprocess


@pytest.fixture
def cache_path():
    return "./pytest_cache/"


@pytest.fixture
def num_topics():
    return 10


def test_demo(cache_path, num_topics):

    topmost.download_dataset("20NG", cache_path=cache_path)

    device = "cuda"  # or "cpu"

    # load a preprocessed dataset
    dataset = topmost.BasicDataset(f"{cache_path}/20NG", device=device, read_labels=True)
    # create a model
    model = topmost.ProdLDA(dataset.vocab_size, num_topics)
    model = model.to(device)

    # create a trainer
    trainer = topmost.BasicTrainer(model, dataset)

    # train the model
    top_words, train_theta = trainer.train()

    # topic diversity and coherence
    TD = eva._diversity(top_words)
    TC = eva._coherence(dataset.train_texts, dataset.vocab, top_words)

    # get doc-topic distributions of testing samples
    test_theta = trainer.test(dataset.test_data)
    # clustering
    clustering_results = eva._clustering(test_theta, dataset.test_labels)
    # classification
    cls_results = eva._cls(train_theta, test_theta, dataset.train_labels, dataset.test_labels)

    new_docs = [
        "This is a new document about space, including words like space, satellite, launch, orbit.",
        "This is a new document about Microsoft Windows, including words like windows, files, dos."
    ]

    preprocess = Preprocess()
    new_parsed_docs, new_bow = preprocess.parse(new_docs, vocab=dataset.vocab)
    new_theta = trainer.test(torch.as_tensor(new_bow.toarray(), device=device).float())



from .basic_dataset import BasicDataset
from .basic_dataset import RawDataset
from .crosslingual_dataset import CrosslingualDataset
from .dynamic_dataset import DynamicDataset

from .download import download_dataset



"""Poincare ball manifold."""

import torch
from .base import Manifold
from .math_util import TanC, ArTanC


class PoincareBall(Manifold):
    """PoicareBall Manifold class.

    We use the following convention: x0^2 + x1^2 + ... + xd^2 < 1 / c. (c < 0)

    So that the Poincare ball radius will be 1 / sqrt(-c).
    Notice that the more close c is to 0, the more flat space will be.
    """

    def __init__(self, ):
        super(PoincareBall, self).__init__()
        self.name = 'PoincareBall'
        self.truncate_c = lambda x: torch.clamp(x, min=-1e5, max=-1e-5)

    def proj(self, x, c):
        c = self.truncate_c(c)
        x_norm = self.clip(x.norm(dim=-1, keepdim=True, p=2))
        max_norm = (1 - self.eps) / c.abs().sqrt()
        cond = x_norm > max_norm
        projected = x / x_norm * max_norm
        return torch.where(cond, projected, x)

    def proj_tan(self, v, x, c):
        return v

    def proj_tan0(self, v, c):
        return v

    def expmap(self, v, x, c):
        c = self.truncate_c(c)
        v_norm = self.clip(v.norm(p=2, dim=-1, keepdim=True))
        second_term = TanC(self._lambda_x(x, c) * v_norm / 2.0, c) * v / v_norm
        gamma = self._mobius_add(x, second_term, c)
        return gamma

    def expmap0(self, v, c):
        c = self.truncate_c(c)
        v_norm = self.clip(v.norm(p=2, dim=-1, keepdim=True))
        gamma = TanC(v_norm, c) * v / v_norm
        return gamma

    def logmap(self, y, x, c):
        c = self.truncate_c(c)
        sub = self._mobius_add(-x, y, c)
        sub_norm = self.clip(sub.norm(p=2, dim=-1, keepdim=True))
        lam = self._lambda_x(x, c)
        return 2.0 / lam * ArTanC(sub_norm, c) * sub / sub_norm

    def logmap0(self, y, c):
        c = self.truncate_c(c)
        y_norm = self.clip(y.norm(p=2, axis=-1, keepdim=True))
        return ArTanC(y_norm, c) * y / y_norm

    def ptransp(self, v, x, y, c):
        c = self.truncate_c(c)
        lambda_x = self._lambda_x(x, c)
        lambda_y = self._lambda_x(y, c)
        return self._gyration(y, -x, v, c) * lambda_x / lambda_y

    def ptransp0(self, v, x, c):
        c = self.truncate_c(c)
        lambda_x = self._lambda_x(x, c)
        return torch.tensor(2.0, dtype=self.dtype) * v / lambda_x

    def dist(self, x, y, c):
        c = self.truncate_c(c)
        return 2.0 * ArTanC(self._mobius_add(-x, y, c).norm(p=2, dim=-1), c)

    def egrad2rgrad(self, grad, x, c):
        c = self.truncate_c(c)
        metric = torch.square(self._lambda_x(x, c))
        return grad / metric

    def inner(self, v1, v2, x, c, keep_shape=False):
        c = self.truncate_c(c)
        metric = torch.square(self._lambda_x(x, c))
        product = v1 * metric * v2
        res = product.sum(dim=-1, keepdim=True)
        if keep_shape:
            # return tf.broadcast_to(res, x.shape)
            last_dim = x.shape.as_list()[-1]
            return torch.cat([res for _ in range(last_dim)], dim=-1)
        return torch.squeeze(res, dim=-1)

    def retraction(self, v, x, c):
        c = self.truncate_c(c)
        new_v = self.expmap(v, x, c)
        return self.proj(new_v, c)

    def _mobius_add(self, x, y, c):
        c = self.truncate_c(c)
        x2 = x.pow(2).sum(dim=-1, keepdim=True)
        y2 = y.pow(2).sum(dim=-1, keepdim=True)
        xy = (x * y).sum(dim=-1, keepdim=True)
        num = (1 - 2 * c * xy - c * y2) * x + (1 + c * x2) * y
        denom = 1 - 2 * c * xy + (c ** 2) * x2 * y2
        return num / self.clip(denom)

    def _mobius_mul(self, x, a, c):
        c = self.truncate_c(c)
        x_norm = self.clip(x.norm(p=2, dim=-1, keepdim=True))
        scale = TanC(a * ArTanC(x_norm, c), c) / x_norm
        return scale * x

    def _mobius_matvec(self, x, a, c):
        c = self.truncate_c(c)
        x_norm = self.clip(x.norm(p=2, dim=-1, keepdim=True))
        mx = torch.matmul(x, a)
        mx_norm = self.clip(mx.norm(p=2, dim=-1, keepdim=True))
        res = TanC(mx_norm / x_norm * ArTanC(x_norm, c), c) * mx / mx_norm
        return res

    def _lambda_x(self, x, c):
        c = self.truncate_c(c)
        x_sqnorm = x.pow(2).sum(dim=-1, keepdim=True)
        return self.clip(2.0 / (1.0 + c * x_sqnorm))

    def _gyration(self, x, y, v, c):
        xy = self._mobius_add(x, y, c)
        yv = self._mobius_add(y, v, c)
        xyv = self._mobius_add(x, yv, c)
        return self._mobius_add(-xy, xyv, c)




import torch
import torch.nn as nn
import torch.nn.functional as F
from . import manifolds
from ..SawETM.SawETM import SawETM


class HyperMiner(SawETM):
    """
        HyperMiner: Topic Taxonomy Mining with Hyperbolic Embedding. NeurIPS 2022.

        Yishi Xu, Dongsheng Wang, Bo Chen, Ruiying Lu, Zhibin Duan, Mingyuan Zhou.

        https://github.com/NoviceStone/HyperMiner
    """
    def __init__(self, vocab_size, num_topics_list, device='cpu', manifold="PoincareBall", clip_r=None, curvature=-0.01, embed_size=50, hidden_size=300, pretrained_WE=None):
        super().__init__(vocab_size, num_topics_list, device, embed_size, hidden_size, pretrained_WE)

        self.manifold = getattr(manifolds, manifold)()

        if curvature is not None:
            self.curvature = torch.tensor([curvature])
            self.curvature = self.curvature.to(device)
        else:
            self.curvature = nn.Parameter(torch.Tensor([-1.]))

        self.clip_r = clip_r

    def feat_clip(self, x):
        x_norm = x.norm(p=2, dim=-1, keepdim=True)
        cond = x_norm > self.clip_r
        projected = x / x_norm * self.clip_r
        return torch.where(cond, projected, x)

    @property
    def bottom_word_embeddings(self):
        return self.rho

    @property
    def topic_embeddings_list(self):
        return self.alpha[::-1]

    def get_phi(self):
        """Returns the factor loading matrix by utilizing sawtooth connection.
        """
        phis = []
        for n in range(self.num_layers):
            if n == 0:
                hyp_rho = self.manifold.proj(
                    self.manifold.expmap0(self.rho, self.curvature), self.curvature)
                hyp_alpha = self.manifold.proj(
                    self.manifold.expmap0(self.alpha[n], self.curvature), self.curvature)
                phi = torch.softmax(-self.manifold.dist(
                    hyp_rho.unsqueeze(1), hyp_alpha.unsqueeze(0), self.curvature), dim=0)
            else:
                hyp_alpha1 = self.manifold.proj(
                    self.manifold.expmap0(self.alpha[n - 1], self.curvature), self.curvature)
                hyp_alpha2 = self.manifold.proj(
                    self.manifold.expmap0(self.alpha[n], self.curvature), self.curvature)
                phi = torch.softmax(-self.manifold.dist(
                    hyp_alpha1.unsqueeze(1).detach(), hyp_alpha2.unsqueeze(0), self.curvature), dim=0)
            phis.append(phi)
        return phis

    def get_beta(self):
        beta_list = list()
        phis = self.get_phi()
        last_beta  = None

        for layer_id, phi in enumerate(phis):
            if layer_id == 0:
                last_beta = phi.T
            else:
                last_beta = torch.matmul(phi.T, last_beta)
            beta_list.append(last_beta)

        beta_list = beta_list[::-1]
        return beta_list

    def get_phi_list(self):
        phis = self.get_phi()
        phis = phis[1:]
        return [item.T for item in phis][::-1]

    def get_theta(self, x):
        hidden_feats = []
        for n in range(self.num_layers):
            if n == 0:
                hidden_feats.append(self.h_encoder[n](x))
            else:
                hidden_feats.append(self.h_encoder[n](hidden_feats[-1]))

        phis = self.get_phi()

        ks = []
        lambs = []
        thetas = []
        phi_by_theta_list = []
        for n in range(self.num_layers - 1, -1, -1):
            if n == self.num_layers - 1:
                joint_feat = hidden_feats[n]
            else:
                joint_feat = torch.cat((hidden_feats[n], phi_by_theta_list[0]), dim=1)

            k, lamb = torch.chunk(F.softplus(self.q_theta[n](joint_feat)), 2, dim=1)
            k = torch.clamp(k, self.wei_shape_min.item(), self.wei_shape_max.item())
            lamb = torch.clamp(lamb, self.real_min.item())

            if self.training:
                lamb = lamb / torch.exp(torch.lgamma(1 + 1 / k))
                theta = self.reparameterize(k, lamb, sample_num=3) if n == 0 else self.reparameterize(k, lamb)
            else:
                theta = torch.min(lamb, self.theta_max)

            phi_by_theta = torch.mm(theta, phis[n].t())
            phi_by_theta_list.insert(0, phi_by_theta)
            thetas.insert(0, theta)
            lambs.insert(0, lamb)
            ks.insert(0, k)

        if self.training:
            return ks, lambs, phi_by_theta_list, thetas
        else:
            return thetas[::-1]

    def forward(self, x):
        """Forward pass: compute the kl loss and data likelihood.
        """
        ks, lambs, phi_by_theta_list, thetas = self.get_theta(x)

        # =================================================================================
        nll = self.get_nll(x, phi_by_theta_list[0])

        kl_loss = []
        for n in range(self.num_layers):
            if n == self.num_layers - 1:
                kl_loss.append(self.kl_weibull_gamma(
                    ks[n], lambs[n], self.gam_prior, self.gam_prior))
            else:
                kl_loss.append(self.kl_weibull_gamma(
                    ks[n], lambs[n], phi_by_theta_list[n + 1], self.gam_prior))

        nelbo = nll + sum(kl_loss)

        return {'loss': nelbo}




import torch
import torch.nn as nn
import torch.nn.functional as F
from .TopicDistQuant import TopicDistQuant
from .TSC import TSC


class TSCTM(nn.Module):
    '''
        Mitigating Data Sparsity for Short Text Topic Modeling by Topic-Semantic Contrastive Learning. EMNLP 2022

        Xiaobao Wu, Anh Tuan Luu, Xinshuai Dong.

        Note: This implementation does not include TSCTM with augmentations. For augmentations, see https://github.com/BobXWu/TSCTM.
    '''

    def __init__(self, vocab_size, num_topics=50, en_units=200, temperature=0.5, weight_contrast=1.0):
        super().__init__()

        self.fc11 = nn.Linear(vocab_size, en_units)
        self.fc12 = nn.Linear(en_units, en_units)
        self.fc21 = nn.Linear(en_units, num_topics)

        self.mean_bn = nn.BatchNorm1d(num_topics)
        self.mean_bn.weight.requires_grad = False
        self.decoder_bn = nn.BatchNorm1d(vocab_size)
        self.decoder_bn.weight.requires_grad = False

        self.fcd1 = nn.Linear(num_topics, vocab_size, bias=False)

        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

        self.topic_dist_quant = TopicDistQuant(num_topics, num_topics)
        self.contrast_loss = TSC(temperature, weight_contrast)

    def get_beta(self):
        return self.fcd1.weight.T

    def encode(self, inputs):
        e1 = F.softplus(self.fc11(inputs))
        e1 = F.softplus(self.fc12(e1))
        return self.mean_bn(self.fc21(e1))

    def decode(self, theta):
        d1 = F.softmax(self.decoder_bn(self.fcd1(theta)), dim=1)
        return d1

    def get_theta(self, inputs):
        theta = self.encode(inputs)
        softmax_theta = F.softmax(theta, dim=1)
        return softmax_theta

    def forward(self, inputs):
        theta = self.encode(inputs)
        softmax_theta = F.softmax(theta, dim=1)

        quant_rst = self.topic_dist_quant(softmax_theta)

        recon = self.decode(quant_rst['quantized'])
        loss = self.loss_function(recon, inputs) + quant_rst['loss']

        features = torch.cat([F.normalize(theta, dim=1).unsqueeze(1)], dim=1)
        contrastive_loss = self.contrast_loss(features, quant_idx=quant_rst['encoding_indices'])
        loss += contrastive_loss

        return {'loss': loss, 'contrastive_loss': contrastive_loss}

    def loss_function(self, recon_x, x):
        loss = -(x * (recon_x).log()).sum(axis=1)
        loss = loss.mean()
        return loss



"""Base manifold."""

import torch


class Manifold(object):
    """Abstract class to define basic operations on a manifold.

    Attributes:
        clip (function): Clips tensor values to a specified min and max.
        dtype: The type of the variables.
        eps (float): A small constant value.
        max_norm (float): The maximum value for number clipping.
        min_norm (float): The minimum value for number clipping.
    """

    def __init__(self, **kwargs):
        """Initialize a manifold.
        """
        super(Manifold, self).__init__()

        self.min_norm = 1e-15
        self.max_norm = 1e15
        self.eps = 1e-5

        self.dtype = kwargs["dtype"] if "dtype" in kwargs else torch.float32
        self.clip = lambda x: torch.clamp(x, min=self.min_norm, max=self.max_norm)

    def proj(self, x, c):
        """A projection function that prevents x from leaving the manifold.
        Args:
            x (tensor): A point should be on the manifold, but it may not meet the manifold constraints.
            c (tensor): The manifold curvature.
        Returns:
            tensor: A projected point, meeting the manifold constraints.
        """
        raise NotImplementedError

    def proj_tan(self, v, x, c):
        """A projection function that prevents v from leaving the tangent space of point x.
        Args:
            v (tensor): A point should be on the tangent space, but it may not meet the manifold constraints.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: A projected point, meeting the tangent space constraints.
        """
        raise NotImplementedError

    def proj_tan0(self, v, c):
        """A projection function that prevents v from leaving the tangent space of origin point.
        Args:
            v (tensor): A point should be on the tangent space, but it may not meet the manifold constraints.
            c (tensor): The manifold curvature.
        Returns:
            tensor: A projected point, meeting the tangent space constraints.
        """
        raise NotImplementedError

    def expmap(self, v, x, c):
        """Map a point v in the tangent space of point x to the manifold.
        Args:
            v (tensor): A point in the tangent space of point x.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of mapping tangent point v to the manifold.
        """
        raise NotImplementedError

    def expmap0(self, v, c):
        """Map a point v in the tangent space of origin point to the manifold.
        Args:
            v (tensor): A point in the tangent space of origin point.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of mapping tangent point v to the manifold.
        """
        raise NotImplementedError

    def logmap(self, y, x, c):
        """Map a point y on the manifold to the tangent space of x.
        Args:
            y (tensor): A point on the manifold.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of mapping y to the tangent space of x.
        """
        raise NotImplementedError

    def logmap0(self, y, c):
        """Map a point y on the manifold to the tangent space of origin point.
        Args:
            y (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of mapping y to the tangent space of origin point.
        """
        raise NotImplementedError

    def ptransp(self, v, x, y, c):
        """Parallel transport function, used to move point v in the tangent space of x to the tangent space of y.
        Args:
            v (tensor): A point in the tangent space of x.
            x (tensor): A point on the manifold.
            y (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of transporting v from the tangent space at x to the tangent space at y.
        """
        raise NotImplementedError

    def ptransp0(self, v, x, c):
        """Parallel transport function, used to move point v in the tangent space of origin point to the tangent space of y.
        Args:
            v (tensor): A point in the tangent space of origin point.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of transporting v from the tangent space at origin point to the tangent space at y.
        """
        raise NotImplementedError

    def dist(self, x, y, c):
        """Calculate the squared geodesic/distance between x and y.
        Args:
            x (tensor): A point on the manifold.
            y (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: the geodesic/distance between x and y.
        """
        raise NotImplementedError

    def egrad2rgrad(self, grad, x, c):
        """Computes Riemannian gradient from the Euclidean gradient, typically used in Riemannian optimizers.
        Args:
            grad (tensor): Euclidean gradient at x.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: Riemannian gradient at x.
        """
        raise NotImplementedError

    def inner(self, v1, v2, x, c, keep_shape):
        """Computes the inner product of a pair of tangent vectors v1 and v2 at x.
        Args:
            v1 (tensor): A tangent point at x.
            v2 (tensor): A tangent point at x.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
            keep_shape (bool, optional): Whether the output tensor keeps shape or not.
        Returns:
            tensor: The inner product of v1 and v2 at x.
        """
        raise NotImplementedError

    def retraction(self, v, x, c):
        """Retraction is a continuous map function from tangent space to the manifold, typically used in Riemannian optimizers.
        The exp map is one of retraction functions.
        Args:
            v (tensor): A tangent point at x.
            x (tensor): A point on the manifold.
            c (tensor): The manifold curvature.
        Returns:
            tensor: The result of mapping tangent point v at x to the manifold.
        """
        return self.proj(self.expmap(v, x, c), c)



import os
from sklearn.datasets import fetch_20newsgroups
from topmost.data import file_utils


def download_save(output_dir):
    names = ['talk.religion.misc', 'comp.windows.x', 'rec.sport.baseball', 'talk.politics.mideast', 'comp.sys.mac.hardware', 'sci.space', 'talk.politics.guns', 'comp.graphics', 'comp.os.ms-windows.misc', 'soc.religion.christian', 'talk.politics.misc', 'rec.motorcycles', 'comp.sys.ibm.pc.hardware', 'rec.sport.hockey', 'misc.forsale', 'sci.crypt', 'rec.autos', 'sci.med', 'sci.electronics', 'alt.atheism']

    # prefixes = set([name.split('.')[0] for name in names])

    groups = {'20ng_all': names}
    # for prefix in prefixes:
    #     elements = [name for name in names if name.split('.')[0] == prefix]
    #     groups['20ng_' + prefix] = elements

    print(groups)

    for group in groups.keys():
        if len(groups[group]) > 1:
            for subset in ['train', 'test', 'all']:
                download_articles(group, groups[group], subset, output_dir)


def download_articles(name, categories, subset, output_dir):
    print("name: ", name)
    print("categories: ", categories)
    print("subset: ", subset)

    data = []
    print("Downloading articles")
    newsgroups_data = fetch_20newsgroups(subset=subset, categories=categories, remove=())

    for i in range(len(newsgroups_data['data'])):
        line = newsgroups_data['data'][i]
        data.append({'text': line, 'group': newsgroups_data['target_names'][newsgroups_data['target'][i]]})

    print("data size: ", len(data))
    print("Saving to", output_dir)
    file_utils.make_dir(output_dir)
    file_utils.save_jsonlist(data, f"{output_dir}/{subset}.jsonlist")



import torch
from torch import nn
from . import utils


class TPD(nn.Module):
    def __init__(self, sinkhorn_alpha, sinkhorn_max_iter=5000, stopThr=.5e-2):
        super().__init__()

        self.sinkhorn_alpha = sinkhorn_alpha
        self.sinkhorn_max_iter = sinkhorn_max_iter
        self.stopThr = stopThr
        self.epsilon = 1e-16

    def forward(self, topic_embeddings_list, weight_loss_TPD=20.0):
        all_loss_TPD = 0.
        transp_list = list()

        num_layers = len(topic_embeddings_list)

        for layer_id in range(num_layers)[:-1]:
            topic_embeddings = topic_embeddings_list[layer_id]
            next_topic_embeddings = topic_embeddings_list[layer_id + 1]
            cost = utils.pairwise_euclidean_distance(topic_embeddings, next_topic_embeddings)
            loss_TPD, transp = self.sinkhorn(cost, return_transp=True)

            all_loss_TPD += loss_TPD
            transp_list.append(transp)

        all_loss_TPD *= weight_loss_TPD / (num_layers - 1)

        return loss_TPD, transp_list

    def sinkhorn(self, M, return_transp=False):
        device = M.device

        a = (torch.ones(M.shape[0]) / M.shape[0]).unsqueeze(1).to(device)
        b = (torch.ones(M.shape[1]) / M.shape[1]).unsqueeze(1).to(device)

        u = (torch.ones_like(a) / a.size()[0]).to(device) # Kx1

        K = torch.exp(-M * self.sinkhorn_alpha)
        err = 1
        cpt = 0
        while err > self.stopThr and cpt < self.sinkhorn_max_iter:
            v = torch.div(b, torch.matmul(K.t(), u) + self.epsilon)
            u = torch.div(a, torch.matmul(K, v) + self.epsilon)
            cpt += 1
            if cpt % 50 == 1:
                bb = torch.mul(v, torch.matmul(K.t(), u))
                err = torch.norm(torch.sum(torch.abs(bb - b), dim=0), p=float('inf'))

        transp = u * (K * v.T)

        loss = torch.sum(transp * M)

        if return_transp:
            return loss, transp

        return loss



"""Euclidean manifold."""

import torch
from .base import Manifold


class Euclidean(Manifold):
    """Euclidean Manifold class. Usually we refer it as R^n.

    Attributes:
        name (str): The manifold name, and its value is "Euclidean".
    """

    def __init__(self, **kwargs):
        """Initialize an Euclidean manifold.
        Args:
            **kwargs: Description
        """
        super(Euclidean, self).__init__(**kwargs)
        self.name = 'Euclidean'

    def proj(self, x, c):
        return x

    def proj_tan(self, v, x, c):
        return v

    def proj_tan0(self, v, c):
        return v

    def expmap(self, v, x, c):
        return v + x

    def expmap0(self, v, c):
        return v

    def logmap(self, y, x, c):
        return y - x

    def logmap0(self, y, c):
        return y

    def ptransp(self, v, x, y, c):
        return torch.ones_like(x) * v

    def ptransp0(self, v, x, c):
        return torch.ones_like(x) * v

    def dist(self, x, y, c):
        sqdis = torch.sum((x - y).pow(2), dim=-1)
        return sqdis.sqrt()

    def egrad2rgrad(self, grad, x, c):
        return grad

    def inner(self, v1, v2, x, c, keep_shape=False):
        if keep_shape:
            # In order to keep the same computation logic in Ada* Optimizer
            return v1 * v2
        else:
            return torch.sum(v1 * v2, dim=-1, keepdim=False)



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class CombinedTM(nn.Module):
    def __init__(self, vocab_size, contextual_embed_size, num_topics=50, en_units=200, dropout=0.4):
        super().__init__()

        self.vocab_size = vocab_size
        self.num_topics = num_topics

        self.a = 1 * np.ones((1, num_topics)).astype(np.float32)
        self.mu2 = nn.Parameter(torch.as_tensor((np.log(self.a).T - np.mean(np.log(self.a), 1)).T))
        self.var2 = nn.Parameter(torch.as_tensor((((1.0 / self.a) * (1 - (2.0 / num_topics))).T + (1.0 / (num_topics * num_topics)) * np.sum(1.0 / self.a, 1)).T))

        self.mu2.requires_grad = False
        self.var2.requires_grad = False

        self.fc_contextual = nn.Linear(contextual_embed_size, vocab_size)
        # self.fc11 = nn.Linear(vocab_size + vocab_size, en_units)
        self.fc11 = nn.Linear(vocab_size, en_units)
        self.fc12 = nn.Linear(en_units, en_units)
        self.fc21 = nn.Linear(en_units, num_topics)
        self.fc22 = nn.Linear(en_units, num_topics)

        # align with the default parameters of tf.contrib.layers.batch_norm
        # https://www.tensorflow.org/versions/r1.15/api_docs/python/tf/contrib/layers/batch_norm
        # center=True (add bias(beta)), scale=False (weight(gamma) is not used)
        self.mean_bn = nn.BatchNorm1d(num_topics, eps=0.001, momentum=0.001, affine=True)
        self.mean_bn.weight.data.copy_(torch.ones(num_topics))
        self.mean_bn.weight.requires_grad = False

        self.logvar_bn = nn.BatchNorm1d(num_topics, eps=0.001, momentum=0.001, affine=True)
        self.logvar_bn.weight.data.copy_(torch.ones(num_topics))
        self.logvar_bn.weight.requires_grad = False

        self.decoder_bn = nn.BatchNorm1d(vocab_size, eps=0.001, momentum=0.001, affine=True)
        self.decoder_bn.weight.data.copy_(torch.ones(vocab_size))
        self.decoder_bn.weight.requires_grad = False

        self.fc1_drop = nn.Dropout(dropout)
        self.theta_drop = nn.Dropout(dropout)

        self.fcd1 = nn.Linear(num_topics, vocab_size, bias=False)
        nn.init.xavier_uniform_(self.fcd1.weight)

    def get_beta(self):
        return self.fcd1.weight.T

    def get_theta(self, x):
        contextual = self.fc_contextual(x[:, self.vocab_size:])
        # combined = torch.cat((x[:, :self.vocab_size], contextual), dim=1)
        combined = contextual

        mu, logvar = self.encode(combined)
        z = self.reparameterize(mu, logvar)
        theta = F.softmax(z, dim=1)
        theta = self.theta_drop(theta)
        if self.training:
            return theta, mu, logvar
        else:
            return theta

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return mu + (eps * std)
        else:
            return mu

    def encode(self, x):
        e1 = F.softplus(self.fc11(x))
        e1 = F.softplus(self.fc12(e1))
        e1 = self.fc1_drop(e1)
        return self.mean_bn(self.fc21(e1)), self.logvar_bn(self.fc22(e1))

    def decode(self, theta):
        d1 = F.softmax(self.decoder_bn(self.fcd1(theta)), dim=1)
        return d1

    def forward(self, x):
        theta, mu, logvar = self.get_theta(x)
        recon_x = self.decode(theta)
        loss = self.loss_function(x[:, :self.vocab_size], recon_x, mu, logvar)
        return {'loss': loss}

    def loss_function(self, x, recon_x, mu, logvar):
        recon_loss = -(x * (recon_x + 1e-10).log()).sum(axis=1)
        var = logvar.exp()
        var_division = var / self.var2
        diff = mu - self.mu2
        diff_term = diff * diff / self.var2
        logvar_division = self.var2.log() - logvar
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(axis=1) - self.num_topics)
        loss = (recon_loss + KLD).mean()
        return loss



"""Math utils functions."""

import torch
from torch import tan, atan, cos, acos, sin, asin
from torch import tanh, atanh, cosh, acosh, sinh, asinh


def Tan(x):
    """Computes tangent of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor. Has the same type as x.
    """
    return tan(x)


def Tanh(x):
    """Computes hyperbolic tangent of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type as x.
    """
    return tanh(torch.clamp(x, min=-15, max=15))


def TanC(x, c):
    """A unified tangent and inverse tangent function for different signs of curvatures.

    This function is used in k-Stereographic model, a unification of constant curvature manifolds.
    Please refer to https://arxiv.org/abs/2007.07698 for more details.

    First-order expansion is used in order to calculate gradients correctly when c is zero.

    Args:
        x (tensor): A tensor.
        c (tensor): Manifold curvature.

    Returns:
        A tensor: Has the same type of x.
    """
    return 1 / c.abs().sqrt() * Tanh(x * c.abs().sqrt())  # c < 0


def ArTan(x):
    """Computes inverse tangent of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type as x.
    """
    return atan(torch.clamp(x, min=-15, max=15))


def ArTanh(x):
    """Computes inverse hyperbolic tangent of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type as x.
    """
    return atanh(torch.clamp(x, min=-1 + 1e-7, max=1 - 1e-7))


def ArTanC(x, c):
    """A unified hyperbolic tangent and inverse hyperbolic tangent function for different signs of curvatures.

    This function is used in k-Stereographic model, a unification of constant curvature manifolds.
    Please refer to https://arxiv.org/abs/2007.07698 for more details.

    First-order expansion is used in order to calculate gradients correctly when c is zero.

    Args:
        x (tensor): A tensor.
        c (tensor): Manifold curvature.

    Returns:
        A tensor: Has the same type of x.
    """
    return 1 / c.abs().sqrt() * ArTanh(x * c.abs().sqrt())  # c < 0


def Cos(x):
    """Computes cosine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return cos(x)


def Cosh(x):
    """Computes hyperbolic cosine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return cosh(torch.clamp(x, min=-15, max=15))


def ArCos(x):
    """Computes inverse cosine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return acos(torch.clamp(x, min=-1 + 1e-7, max=1 - 1e-7))


def ArCosh(x):
    """Computes inverse hyperbolic cosine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return acosh(torch.clamp(x, min=1 + 1e-7, max=1e15))


def Sin(x):
    """Computes sine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return sin(x)


def Sinh(x):
    """Computes hyperbolic sine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return sinh(torch.clamp(x, min=-15, max=15))


def SinC(x, c):
    """A unified sine and inverse sine function for different signs of curvatures.

    This function is used in k-Stereographic model, a unification of constant curvature manifolds.
    Please refer to https://arxiv.org/abs/2007.07698 for more details.

    First-order expansion is used in order to calculate gradients correctly when c is zero.

    Args:
        x (tensor): A tensor.
        c (tensor): Manifold curvature.

    Returns:
        A tensor: Has the same type of x.
    """
    return 1 / c.abs().sqrt() * Sinh(x * c.abs().sqrt())  # c < 0


def ArSin(x):
    """Computes inverse sine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return asin(torch.clamp(x, min=-1 + 1e-7, max=1 - 1e-7))


def ArSinh(x):
    """Computes inverse hyperbolic sine of x element-wise.

    Args:
        x (tensor): A tensor.

    Returns:
        A tensor: Has the same type of x.
    """
    return asinh(torch.clamp(x, min=-15, max=15))


def ArSinC(x, c):
    """A unified hyperbolic sine and inverse hyperbolic sine function for different signs of curvatures.

    This function is used in k-Stereographic model, a unification of constant curvature manifolds.
    Please refer to https://arxiv.org/abs/2007.07698 for more details.

    First-order expansion is used in order to calculate gradients correctly when c is zero.

    Args:
        x (tensor): A tensor.
        c (tensor): Manifold curvature.

    Returns:
        A tensor: Has the same type of x.
    """
    return 1 / c.abs().sqrt() * ArSinh(x * c.abs().sqrt())  # c < 0



import os
import numpy as np
import scipy
import torch
from collections import defaultdict
from torch.utils.data import Dataset, DataLoader
from . import file_utils


class _BilingualDataset(Dataset):
    def __init__(self, bow_en, bow_cn):
        self.bow_en = bow_en
        self.bow_cn = bow_cn
        self.bow_size_en = len(self.bow_en)
        self.bow_size_cn = len(self.bow_cn)

    def __len__(self):
        return max(self.bow_size_en, self.bow_size_cn)

    def __getitem__(self, index):
        return_dict = {
            'bow_en': self.bow_en[(index % self.bow_size_en)],
            'bow_cn': self.bow_cn[(index % self.bow_size_cn)]
        }
        return return_dict


class CrosslingualDataset:
    def __init__(self, dataset_dir, lang1, lang2, dict_path, device='cpu', batch_size=200, as_tensor=True):
        self.batch_size = batch_size

        self.train_texts_en, self.test_texts_en, self.train_bow_en, self.test_bow_en, self.train_labels_en, self.test_labels_en, self.vocab_en, self.word2id_en, self.id2word_en = self.read_data(dataset_dir, lang=lang1)
        self.train_texts_cn, self.test_texts_cn, self.train_bow_cn, self.test_bow_cn, self.train_labels_cn, self.test_labels_cn, self.vocab_cn, self.word2id_cn, self.id2word_cn = self.read_data(dataset_dir, lang=lang2)

        self.train_size_en = len(self.train_texts_en)
        self.train_size_cn = len(self.train_texts_cn)
        self.vocab_size_en = len(self.vocab_en)
        self.vocab_size_cn = len(self.vocab_cn)

        self.trans_dict, self.trans_matrix_en, self.trans_matrix_cn = self.parse_dictionary(dict_path)

        self.pretrained_WE_en = scipy.sparse.load_npz(os.path.join(dataset_dir, f'word2vec_{lang1}.npz')).toarray()
        self.pretrained_WE_cn = scipy.sparse.load_npz(os.path.join(dataset_dir, f'word2vec_{lang2}.npz')).toarray()

        self.Map_en2cn = self.get_Map(self.trans_matrix_en, self.train_bow_en)
        self.Map_cn2en = self.get_Map(self.trans_matrix_cn, self.train_bow_cn)

        if as_tensor:
            self.train_bow_en = self.move_to_device(self.train_bow_en, device)
            self.test_bow_en = self.move_to_device(self.test_bow_en, device)
            self.train_bow_cn = self.move_to_device(self.train_bow_cn, device)
            self.test_bow_cn = self.move_to_device(self.test_bow_cn, device)

            self.train_dataloader = DataLoader(_BilingualDataset(self.train_bow_en, self.train_bow_cn), batch_size=batch_size, shuffle=True)
            self.test_dataloader = DataLoader(_BilingualDataset(self.test_bow_en, self.test_bow_cn), batch_size=batch_size, shuffle=False)

    def move_to_device(self, bow, device):
        return torch.as_tensor(bow, device=device).float()

    def read_data(self, dataset_dir, lang):
        train_texts = file_utils.read_text(os.path.join(dataset_dir, 'train_texts_{}.txt'.format(lang)))
        test_texts = file_utils.read_text(os.path.join(dataset_dir, 'test_texts_{}.txt'.format(lang)))
        vocab = file_utils.read_text(os.path.join(dataset_dir, 'vocab_{}'.format(lang)))
        word2id = dict(zip(vocab, range(len(vocab))))
        id2word = dict(zip(range(len(vocab)), vocab))

        train_bow = scipy.sparse.load_npz(os.path.join(dataset_dir, 'train_bow_matrix_{}.npz'.format(lang))).toarray()
        test_bow = scipy.sparse.load_npz(os.path.join(dataset_dir, 'test_bow_matrix_{}.npz'.format(lang))).toarray()

        train_labels = np.loadtxt(f'{dataset_dir}/train_labels_{lang}.txt').astype('int32')
        test_labels = np.loadtxt(f'{dataset_dir}/test_labels_{lang}.txt').astype('int32')

        return train_texts, test_texts, train_bow, test_bow, train_labels, test_labels, vocab, word2id, id2word

    def parse_dictionary(self, dict_path):
        trans_dict = defaultdict(set)

        trans_matrix_en = np.zeros((self.vocab_size_en, self.vocab_size_cn), dtype='int32')
        trans_matrix_cn = np.zeros((self.vocab_size_cn, self.vocab_size_en), dtype='int32')

        dict_texts = file_utils.read_text(dict_path)

        for line in dict_texts:
            terms = (line.strip()).split()
            if len(terms) == 2:
                cn_term = terms[0]
                en_term = terms[1]
                if cn_term in self.word2id_cn and en_term in self.word2id_en:
                    trans_dict[cn_term].add(en_term)
                    trans_dict[en_term].add(cn_term)
                    cn_term_id = self.word2id_cn[cn_term]
                    en_term_id = self.word2id_en[en_term]

                    trans_matrix_en[en_term_id][cn_term_id] = 1
                    trans_matrix_cn[cn_term_id][en_term_id] = 1

        return trans_dict, trans_matrix_en, trans_matrix_cn

    def get_Map(self, trans_matrix, bow):
        Map = (trans_matrix * bow.sum(0)[:, np.newaxis]).astype('float32')
        Map = Map + 1
        Map_sum = np.sum(Map, axis=1)
        t_index = Map_sum > 0
        Map[t_index, :] = Map[t_index, :] / Map_sum[t_index, np.newaxis]

        return Map



import pytest

import sys
sys.path.append('../')

import topmost
from topmost import download_dataset
from topmost import BasicDataset, HierarchicalTrainer


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics_list():
    return [2, 4, 8]


def hierarchical_model_test(model_module, dataset, num_topics_list):
    model = model_module(num_topics_list=num_topics_list, vocab_size=dataset.vocab_size)
    trainer = HierarchicalTrainer(model, dataset, verbose=True, epochs=1)
    trainer.train()

    beta = trainer.get_beta()
    assert len(beta) == len(num_topics_list)

    for i, layer_beta in enumerate(beta):
        assert layer_beta.shape == (num_topics_list[i], dataset.vocab_size)

    train_theta, test_theta = trainer.export_theta()
    assert len(train_theta) == len(num_topics_list)
    assert len(test_theta) == len(num_topics_list)

    for i, layer_theta in enumerate(train_theta):
        assert layer_theta.shape == (len(dataset.train_texts), num_topics_list[i])

    for i, layer_theta in enumerate(test_theta):
        assert layer_theta.shape == (len(dataset.test_texts), num_topics_list[i])


def test_models(cache_path, num_topics_list):
    download_dataset("20NG", cache_path=cache_path)
    dataset = BasicDataset(f"{cache_path}/20NG", as_tensor=True)

    model_info = [
        topmost.SawETM,
        topmost.HyperMiner,
        topmost.TraCo,
    ]

    for model_module in model_info:
        print(model_module)
        hierarchical_model_test(model_module, dataset, num_topics_list)



import torch
import torch.nn as nn


# Topic-Semantic Contrastive Learning
class TSC(nn.Module):
    def __init__(self, temperature=0.07, weight_contrast=None, use_aug=False):
        super().__init__()
        self.use_aug = use_aug
        self.temperature = temperature
        self.weight_contrast = weight_contrast

    def forward(self, features, quant_idx=None, weight_same_quant=None):
        device = features.device

        batch_size = features.shape[0]
        mask = torch.eye(batch_size, dtype=torch.float32).to(device)

        contrast_count = features.shape[1]
        contrast_feature = torch.cat(torch.unbind(features, dim=1), dim=0)
        anchor_feature = contrast_feature
        anchor_count = contrast_count

        anchor_dot_contrast = torch.div(
            torch.matmul(anchor_feature, contrast_feature.T),
            self.temperature
        )

        # for numerical stability
        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        # tile mask
        mask = mask.repeat(anchor_count, contrast_count)
        # mask-out self-contrast cases.
        # logits_mask is 1 - eye matrix
        logits_mask = torch.scatter(
            torch.ones_like(mask),
            1,
            torch.arange(batch_size * anchor_count).view(-1, 1).to(device),
            0
        )

        mask = mask * logits_mask

        t_quant_idx = quant_idx.contiguous().view(-1, 1)

        # quant_idx_mask: 1 means same quantization; 0 means different quantization
        quant_idx_mask = torch.eq(t_quant_idx, t_quant_idx.T).float()
        quant_idx_mask = quant_idx_mask.repeat(anchor_count, contrast_count)

        exp_logits = torch.exp(logits) * (1 - quant_idx_mask)
        sum_exp_logits = exp_logits.sum(1, keepdim=True)

        if not self.use_aug:
            # quant_idx_mask includes self-contrast cases.
            # logits * logits_mask is to remove the positive pair but keep the negative pair in the self-contrast cases.
            # This is because some samples do not have positive pairs.
            log_prob = logits * logits_mask - torch.log(sum_exp_logits + 1e-10)
            mean_log_prob_pos = (quant_idx_mask * log_prob).sum(1) / quant_idx_mask.sum(1)
        else:
            log_prob = logits - torch.log(sum_exp_logits + 1e-10)
            # between original and augmented samples.
            mean_log_prob_pos = (mask * log_prob).sum(1) / mask.sum(1)

            # between original samples.
            same_quant_mask = quant_idx_mask * logits_mask
            same_quant_mean_log_prob_pos = (same_quant_mask * log_prob).sum(1) / (same_quant_mask.sum(1) + 1e-10)
            mean_log_prob_pos += weight_same_quant * same_quant_mean_log_prob_pos

        loss = - self.weight_contrast * mean_log_prob_pos
        loss = loss.view(anchor_count, batch_size).sum(axis=0).mean()

        return loss





import torch
import torch.nn as nn
import torch.nn.functional as F


class ETM(nn.Module):
    '''
        Topic Modeling in Embedding Spaces. TACL 2020

        Adji B. Dieng, Francisco J. R. Ruiz, David M. Blei.
    '''
    def __init__(self, vocab_size, embed_size=200, num_topics=50, en_units=800, dropout=0., pretrained_WE=None, train_WE=False):
        super().__init__()

        if pretrained_WE is not None:
            self.word_embeddings = nn.Parameter(torch.from_numpy(pretrained_WE).float())
        else:
            self.word_embeddings = nn.Parameter(torch.randn((vocab_size, embed_size)))

        self.word_embeddings.requires_grad = train_WE

        self.topic_embeddings = nn.Parameter(torch.randn((num_topics, self.word_embeddings.shape[1])))

        self.encoder1 = nn.Sequential(
            nn.Linear(vocab_size, en_units),
            nn.ReLU(),
            nn.Linear(en_units, en_units),
            nn.ReLU(),
            nn.Dropout(dropout)
        )

        self.fc21 = nn.Linear(en_units, num_topics)
        self.fc22 = nn.Linear(en_units, num_topics)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return mu + (eps * std)
        else:
            return mu

    def encode(self, x):
        e1 = self.encoder1(x)
        return self.fc21(e1), self.fc22(e1)

    def get_theta(self, x):
        # Warn: normalize the input if use Relu.
        # https://github.com/adjidieng/ETM/issues/3
        norm_x = x / x.sum(1, keepdim=True)
        mu, logvar = self.encode(norm_x)
        z = self.reparameterize(mu, logvar)
        theta = F.softmax(z, dim=-1)
        if self.training:
            return theta, mu, logvar
        else:
            return theta

    def get_beta(self):
        beta = F.softmax(torch.matmul(self.topic_embeddings, self.word_embeddings.T), dim=1)
        return beta

    def forward(self, x, avg_loss=True):
        theta, mu, logvar = self.get_theta(x)
        beta = self.get_beta()
        recon_x = torch.matmul(theta, beta)

        loss = self.loss_function(x, recon_x, mu, logvar, avg_loss)
        return {'loss': loss}

    def loss_function(self, x, recon_x, mu, logvar, avg_loss=True):
        recon_loss = -(x * (recon_x + 1e-12).log()).sum(1)
        KLD = -0.5 * (1 + logvar - mu ** 2 - logvar.exp()).sum(1)
        loss = (recon_loss + KLD)

        if avg_loss:
            loss = loss.mean()

        return loss



'''
This script is partially based on https://github.com/dallascard/scholar.
'''

import os
import re
import string
import gensim.downloader
from collections import Counter
import numpy as np
import scipy.sparse
from tqdm import tqdm
from sklearn.feature_extraction.text import CountVectorizer

from topmost.data import file_utils
from topmost.utils._utils import get_stopwords_set
from topmost.utils.logger import Logger


logger = Logger("WARNING")


# compile some regexes
punct_chars = list(set(string.punctuation) - set("'"))
punct_chars.sort()
punctuation = ''.join(punct_chars)
replace = re.compile('[%s]' % re.escape(punctuation))
alpha = re.compile('^[a-zA-Z_]+$')
alpha_or_num = re.compile('^[a-zA-Z_]+|[0-9_]+$')
alphanum = re.compile('^[a-zA-Z0-9_]+$')


class Tokenizer:
    def __init__(self,
                 stopwords,
                 keep_num,
                 keep_alphanum,
                 strip_html,
                 no_lower,
                 min_length
                ):
        self.keep_num = keep_num
        self.keep_alphanum = keep_alphanum
        self.strip_html = strip_html
        self.lower = not no_lower
        self.min_length = min_length

        self.stopword_set = get_stopwords_set(stopwords)

    def clean_text(self, text, strip_html=False, lower=True, keep_emails=False, keep_at_mentions=False):
        # remove html tags
        if strip_html:
            text = re.sub(r'<[^>]+>', '', text)
        else:
            # replace angle brackets
            text = re.sub(r'<', '(', text)
            text = re.sub(r'>', ')', text)
        # lower case
        if lower:
            text = text.lower()
        # eliminate email addresses
        if not keep_emails:
            text = re.sub(r'\S+@\S+', ' ', text)
        # eliminate @mentions
        if not keep_at_mentions:
            text = re.sub(r'\s@\S+', ' ', text)
        # replace underscores with spaces
        text = re.sub(r'_', ' ', text)
        # break off single quotes at the ends of words
        text = re.sub(r'\s\'', ' ', text)
        text = re.sub(r'\'\s', ' ', text)
        # remove periods
        text = re.sub(r'\.', '', text)
        # replace all other punctuation (except single quotes) with spaces
        text = replace.sub(' ', text)
        # remove single quotes
        text = re.sub(r'\'', '', text)
        # replace all whitespace with a single space
        text = re.sub(r'\s', ' ', text)
        # strip off spaces on either end
        text = text.strip()
        return text

    def tokenize(self, text):
        text = self.clean_text(text, self.strip_html, self.lower)
        tokens = text.split()

        tokens = ['_' if t in self.stopword_set else t for t in tokens]

        # remove tokens that contain numbers
        if not self.keep_alphanum and not self.keep_num:
            tokens = [t if alpha.match(t) else '_' for t in tokens]

        # or just remove tokens that contain a combination of letters and numbers
        elif not self.keep_alphanum:
            tokens = [t if alpha_or_num.match(t) else '_' for t in tokens]

        # drop short tokens
        if self.min_length > 0:
            tokens = [t if len(t) >= self.min_length else '_' for t in tokens]

        unigrams = [t for t in tokens if t != '_']
        # counts = Counter()
        # counts.update(unigrams)

        return unigrams


def make_word_embeddings(vocab):
    glove_vectors = gensim.downloader.load('glove-wiki-gigaword-200')
    word_embeddings = np.zeros((len(vocab), glove_vectors.vectors.shape[1]))

    num_found = 0

    try:
        key_word_list = glove_vectors.index_to_key
    except:
        key_word_list = glove_vectors.index2word

    for i, word in enumerate(tqdm(vocab, desc="loading word embeddings")):
        if word in key_word_list:
            word_embeddings[i] = glove_vectors[word]
            num_found += 1

    logger.info(f'number of found embeddings: {num_found}/{len(vocab)}')

    return scipy.sparse.csr_matrix(word_embeddings)


class Preprocess:
    def __init__(self,
                 tokenizer=None,
                 test_sample_size=None,
                 test_p=0.2,
                 stopwords="English",
                 min_doc_count=0,
                 max_doc_freq=1.0,
                 keep_num=False,
                 keep_alphanum=False,
                 strip_html=False,
                 no_lower=False,
                 min_length=3,
                 min_term=0,
                 vocab_size=None,
                 seed=42,
                 verbose=True
                ):
        """
        Args:
            test_sample_size:
                Size of the test set.
            test_p:
                Proportion of the test set. This helps sample the train set based on the size of the test set.
            stopwords:
                List of stopwords to exclude.
            min-doc-count:
                Exclude words that occur in less than this number of documents.
            max_doc_freq:
                Exclude words that occur in more than this proportion of documents.
            keep-num:
                Keep tokens made of only numbers.
            keep-alphanum:
                Keep tokens made of a mixture of letters and numbers.
            strip_html:
                Strip HTML tags.
            no-lower:
                Do not lowercase text
            min_length:
                Minimum token length.
            min_term:
                Minimum term number
            vocab-size:
                Size of the vocabulary (by most common in the union of train and test sets, following above exclusions)
            seed:
                Random integer seed (only relevant for choosing test set)
        """

        self.test_sample_size = test_sample_size
        self.min_doc_count = min_doc_count
        self.max_doc_freq = max_doc_freq
        self.min_term = min_term
        self.test_p = test_p
        self.vocab_size = vocab_size
        self.seed = seed

        if tokenizer is not None:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = Tokenizer(stopwords, keep_num, keep_alphanum, strip_html, no_lower, min_length).tokenize

        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

    def parse(self, texts, vocab):
        if not isinstance(texts, list):
            texts = [texts]

        vocab_set = set(vocab)
        parsed_texts = list()
        for i, text in enumerate(tqdm(texts, desc="parsing texts")):
            tokens = self.tokenizer(text)
            tokens = [t for t in tokens if t in vocab_set]
            parsed_texts.append(" ".join(tokens))

        vectorizer = CountVectorizer(vocabulary=vocab, tokenizer=lambda x: x.split())
        sparse_bow = vectorizer.fit_transform(parsed_texts)
        return parsed_texts, sparse_bow

    def preprocess_jsonlist(self, dataset_dir, label_name=None):
        train_items = file_utils.read_jsonlist(os.path.join(dataset_dir, 'train.jsonlist'))
        test_items = file_utils.read_jsonlist(os.path.join(dataset_dir, 'test.jsonlist'))

        logger.info(f"Found training documents {len(train_items)} testing documents {len(test_items)}")

        raw_train_texts = []
        train_labels = []
        raw_test_texts = []
        test_labels = []

        for item in train_items:
            raw_train_texts.append(item['text'])

            if label_name is not None:
                train_labels.append(item[label_name])
 
        for item in test_items:
            raw_test_texts.append(item['text'])

            if label_name is not None:
                test_labels.append(item[label_name])

        rst = self.preprocess(raw_train_texts, train_labels, raw_test_texts, test_labels)

        return rst

    def convert_labels(self, train_labels, test_labels):
        if train_labels is not None:
            label_list = list(set(train_labels).union(set(test_labels)))
            label_list.sort()
            n_labels = len(label_list)
            label2id = dict(zip(label_list, range(n_labels)))

            logger.info(f"label2id: {label2id}")

            train_labels = [label2id[label] for label in train_labels]

            if test_labels is not None:
                test_labels = [label2id[label] for label in test_labels]

        return train_labels, test_labels

    def preprocess(
            self,
            raw_train_texts,
            train_labels=None,
            raw_test_texts=None,
            test_labels=None,
            pretrained_WE=False
        ):
        np.random.seed(self.seed)

        train_texts = list()
        test_texts = list()
        word_counts = Counter()
        doc_counts_counter = Counter()

        train_labels, test_labels = self.convert_labels(train_labels, test_labels)

        for text in tqdm(raw_train_texts, desc="loading train texts"):
            tokens = self.tokenizer(text)
            word_counts.update(tokens)
            doc_counts_counter.update(set(tokens))
            parsed_text = ' '.join(tokens)
            train_texts.append(parsed_text)

        if raw_test_texts:
            for text in tqdm(raw_test_texts, desc="loading test texts"):
                tokens = self.tokenizer(text)
                word_counts.update(tokens)
                doc_counts_counter.update(set(tokens))
                parsed_text = ' '.join(tokens)
                test_texts.append(parsed_text)

        words, doc_counts = zip(*doc_counts_counter.most_common())
        doc_freqs = np.array(doc_counts) / float(len(train_texts) + len(test_texts))

        vocab = [word for i, word in enumerate(words) if doc_counts[i] >= self.min_doc_count and doc_freqs[i] <= self.max_doc_freq]

        # filter vocabulary
        if self.vocab_size is not None:
            vocab = vocab[:self.vocab_size]

        vocab.sort()

        train_idx = [i for i, text in enumerate(train_texts) if len(text.split()) >= self.min_term]
        train_idx = np.asarray(train_idx)

        if raw_test_texts is not None:
            test_idx = [i for i, text in enumerate(test_texts) if len(text.split()) >= self.min_term]
            test_idx = np.asarray(test_idx)

        # randomly sample
        if self.test_sample_size:
            logger.info("sample train and test sets...")

            train_num = len(train_idx)
            test_num = len(test_idx)
            test_sample_size = min(test_num, self.test_sample_size)
            train_sample_size = int((test_sample_size / self.test_p) * (1 - self.test_p))
            if train_sample_size > train_num:
                test_sample_size = int((train_num / (1 - self.test_p)) * self.test_p)
                train_sample_size = train_num

            train_idx = train_idx[np.sort(np.random.choice(train_num, train_sample_size, replace=False))]
            test_idx = test_idx[np.sort(np.random.choice(test_num, test_sample_size, replace=False))]

            logger.info(f"sampled train size: {len(train_idx)}")
            logger.info(f"sampled train size: {len(test_idx)}")

        train_texts, train_bow = self.parse([train_texts[i] for i in train_idx], vocab)

        rst = {
            'vocab': vocab,
            'train_bow': train_bow,
            "train_texts": train_texts
        }

        if train_labels is not None:
            rst['train_labels'] = np.asarray(train_labels)[train_idx]

        logger.info(f"Real vocab size: {len(vocab)}")
        logger.info(f"Real training size: {len(train_texts)} \t avg length: {rst['train_bow'].sum() / len(train_texts):.3f}")

        if raw_test_texts is not None:
            rst['test_texts'], rst['test_bow'] = self.parse(np.asarray(test_texts)[test_idx].tolist(), vocab)

            if test_labels is not None:
                rst['test_labels'] = np.asarray(test_labels)[test_idx]

            logger.info(f"Real testing size: {len(rst['test_texts'])} \t avg length: {rst['test_bow'].sum() / len(rst['test_texts']):.3f}")

        if pretrained_WE:
            rst['word_embeddings'] = make_word_embeddings(vocab)

        return rst

    def save(self, output_dir, vocab, train_texts, train_bow, word_embeddings=None, train_labels=None, test_texts=None, test_bow=None, test_labels=None):
        file_utils.make_dir(output_dir)

        file_utils.save_text(vocab, f"{output_dir}/vocab.txt")
        file_utils.save_text(train_texts, f"{output_dir}/train_texts.txt")
        scipy.sparse.save_npz(f"{output_dir}/train_bow.npz", scipy.sparse.csr_matrix(train_bow))
        if word_embeddings:
            scipy.sparse.save_npz(f"{output_dir}/word_embeddings.npz", word_embeddings)

        if train_labels is not None:
            np.savetxt(f"{output_dir}/train_labels.txt", train_labels, fmt='%i')

        if test_bow is not None:
            scipy.sparse.save_npz(f"{output_dir}/test_bow.npz", scipy.sparse.csr_matrix(test_bow))

        if test_texts is not None:
            file_utils.save_text(test_texts, f"{output_dir}/test_texts.txt")

            if test_labels is not None:
                np.savetxt(f"{output_dir}/test_labels.txt", test_labels, fmt='%i')



import numpy as np
from topmost.data import file_utils


def get_top_words(beta, vocab, num_top_words, verbose=False):
    topic_str_list = list()
    for i, topic_dist in enumerate(beta):
        topic_words = np.array(vocab)[np.argsort(topic_dist)][:-(num_top_words + 1):-1]
        topic_str = ' '.join(topic_words)
        topic_str_list.append(topic_str)
        if verbose:
            print('Topic {}: {}'.format(i, topic_str))

    return topic_str_list


def get_stopwords_set(stopwords=[]):
    from topmost.data import download_dataset

    if stopwords == 'English':
        from gensim.parsing.preprocessing import STOPWORDS as stopwords

    elif stopwords in ['mallet', 'snowball']:
        download_dataset('stopwords', cache_path='./')
        path = f'./stopwords/{stopwords}_stopwords.txt'
        stopwords = file_utils.read_text(path)

    stopword_set = frozenset(stopwords)

    return stopword_set


if __name__ == '__main__':
    print(list(get_stopwords_set('English'))[:10])
    print(list(get_stopwords_set('mallet'))[:10])
    print(list(get_stopwords_set('snowball'))[:10])
    print(list(get_stopwords_set())[:10])



import torch


def sinkhorn_loss(M, a, b, lambda_sh, numItermax=5000, stopThr=.5e-2):
    device = a.device

    u = (torch.ones_like(a) / a.size()[0]).to(device)
    # TODO v is zeros in the tensorflow code.
    # v = (torch.ones_like(b)).to(device)

    K = torch.exp(-M * lambda_sh)
    err = 1
    cpt = 0
    while err > stopThr and cpt < numItermax:
        u = torch.div(a, torch.matmul(K, torch.div(b, torch.matmul(u.t(), K).t()))) 
        cpt += 1
        if cpt % 20 == 1:
            v = torch.div(b, torch.matmul(K.t(), u))  
            u = torch.div(a, torch.matmul(K, v))
            bb = torch.mul(v, torch.matmul(K.t(), u))
            err = torch.norm(torch.sum(torch.abs(bb - b), dim=0), p=float('inf'))

    sinkhorn_divergences = torch.sum(torch.mul(u, torch.matmul(torch.mul(K, M), v)), dim=0)

    return sinkhorn_divergences



import numpy as np


def get_child_topic_idx(phi, topic_idx):
    num_topic, next_num_topic = phi.shape
    num_child_topic = round(next_num_topic / num_topic)

    row = phi[topic_idx]
    child_topic_idx_list = np.argsort(row)[:-(num_child_topic + 1):-1].tolist()

    return child_topic_idx_list


def build_hierarchy(hierarchical_topic_dict, phi_list, topic_idx_list=None, layer_id=0):
    # first layer.
    if topic_idx_list is None:
        topic_idx_list = list(range(phi_list[0].shape[0]))

    # last layer, where layer_id == L-1.
    if layer_id >= len(phi_list):
        # return the topic strings at the last layer.
        hierarchy = np.asarray(hierarchical_topic_dict[layer_id])[topic_idx_list].tolist()
        return hierarchy

    # NOT the last layer.
    hierarchy = dict()
    phi = phi_list[layer_id]

    for topic_idx in topic_idx_list:
        child_topic_idx_list = get_child_topic_idx(phi, topic_idx)
        hierarchy[hierarchical_topic_dict[layer_id][topic_idx]] = build_hierarchy(hierarchical_topic_dict, phi_list, child_topic_idx_list, layer_id + 1)

    return hierarchy



import torch
from torch.optim.lr_scheduler import StepLR
from collections import defaultdict
import numpy as np
from tqdm import tqdm
from topmost.utils import _utils
from topmost.utils.logger import Logger


logger = Logger("WARNING")


class CrosslingualTrainer:
    def __init__(self,
                 model,
                 dataset,
                 num_top_words=15,
                 epochs=500,
                 learning_rate=0.002,
                 batch_size=200,
                 lr_scheduler=None,
                 lr_step_size=125,
                 log_interval=5,
                 verbose=False
                ):

        self.model = model
        self.dataset = dataset
        self.num_top_words = num_top_words
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.lr_scheduler = lr_scheduler
        self.lr_step_size = lr_step_size
        self.log_interval = log_interval

        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

    def make_optimizer(self):
        args_dict = {
            'params': self.model.parameters(),
            'lr': self.learning_rate
        }

        optimizer = torch.optim.Adam(**args_dict)
        return optimizer

    def make_lr_scheduler(self, optimizer):
        if self.lr_scheduler == 'StepLR':
            logger.info("using lr_scheduler")
            lr_scheduler = StepLR(optimizer, step_size=self.lr_step_size, gamma=0.5, verbose=False)
        else:
            raise NotImplementedError(self.lr_scheduler)

        return lr_scheduler

    def train(self):
        data_size = len(self.dataset.train_dataloader.dataset)
        optimizer = self.make_optimizer()

        if self.lr_scheduler:
            lr_scheduler = self.make_lr_scheduler(optimizer)

        for epoch in tqdm(range(1, self.epochs + 1)):


            loss_rst_dict = defaultdict(float)

            self.model.train()
            for batch_data in self.dataset.train_dataloader:
                batch_bow_en = batch_data['bow_en']
                batch_bow_cn = batch_data['bow_cn']
                params_list = [batch_bow_en, batch_bow_cn]

                rst_dict = self.model(*params_list)

                batch_loss = rst_dict['loss']

                for key in rst_dict:
                    if 'loss' in key:
                        loss_rst_dict[key] += rst_dict[key]

                optimizer.zero_grad()
                batch_loss.backward()
                optimizer.step()

            if self.lr_scheduler:
                lr_scheduler.step()

            if epoch % self.log_interval == 0:
                output_log = f'Epoch: {epoch:03d}'
                for key in loss_rst_dict:
                    output_log += f' {key}: {loss_rst_dict[key] / data_size :.3f}'

                logger.info(output_log)

        top_words_en, top_words_cn = self.get_top_words()
        train_theta_en, train_theta_cn = self.test(self.dataset.train_bow_en, self.dataset.train_bow_cn)

        return top_words_en, top_words_cn, train_theta_en, train_theta_cn

    def test(self, bow_en, bow_cn):
        theta_en = self.infer_theta(bow_en, lang='en')
        theta_cn = self.infer_theta(bow_cn, lang='cn')

        return theta_en, theta_cn

    def infer_theta(self, bow, lang):
        theta_list = list()
        data_size = bow.shape[0]
        all_idx = torch.split(torch.arange(data_size,), self.batch_size)
        with torch.no_grad():
            self.model.eval()
            for idx in all_idx:
                batch_bow = bow[idx]
                theta = self.model.get_theta(batch_bow, lang)
                theta_list.extend(theta.detach().cpu().numpy().tolist())

        return np.asarray(theta_list)

    def get_beta(self):
        beta_en, beta_cn = self.model.get_beta()
        beta_en = beta_en.detach().cpu().numpy()
        beta_cn = beta_cn.detach().cpu().numpy()

        return beta_en, beta_cn

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        beta_en, beta_cn = self.get_beta()
        top_words_en = _utils.get_top_words(beta_en, self.dataset.vocab_en, num_top_words)
        top_words_cn = _utils.get_top_words(beta_cn, self.dataset.vocab_cn, num_top_words)

        return top_words_en, top_words_cn

    def export_theta(self):
        train_theta_en, train_theta_cn = self.test(self.dataset.train_bow_en, self.dataset.train_bow_cn)
        test_theta_en, test_theta_cn = self.test(self.dataset.test_bow_en, self.dataset.test_bow_cn)

        return train_theta_en, train_theta_cn, test_theta_en, test_theta_cn



import logging


class Logger:
    def __init__(self, level):
        self.logger = logging.getLogger('TopMost')
        self.set_level(level)
        self._add_handler()
        self.logger.propagate = False

    def info(self, message):
        self.logger.info(f"{message}")

    def warning(self, message):
        self.logger.warning(f"WARNING: {message}")

    def set_level(self, level):
        levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        if level in levels:
            self.logger.setLevel(level)

    def _add_handler(self):
        sh = logging.StreamHandler()
        sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(message)s'))
        self.logger.addHandler(sh)

        # Remove duplicate handlers
        if len(self.logger.handlers) > 1:
            self.logger.handlers = [self.logger.handlers[0]]



import pytest

import sys
sys.path.append('../')

import topmost
from topmost import download_dataset
from topmost import CrosslingualDataset, CrosslingualTrainer


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 5


def crosslingual_model_test(model, dataset, num_topics):

    trainer = CrosslingualTrainer(model, dataset, verbose=True, epochs=1)
    trainer.train()

    beta_en, beta_cn = trainer.get_beta()
    assert beta_en.shape == beta_cn.shape
    assert beta_en.shape == (num_topics, dataset.vocab_size_en)

    train_theta_en, train_theta_cn, test_theta_en, test_theta_cn = trainer.export_theta()
    assert train_theta_en.shape == (len(dataset.train_texts_en), num_topics)
    assert test_theta_en.shape == (len(dataset.test_texts_en), num_topics)
    assert train_theta_cn.shape == (len(dataset.train_texts_cn), num_topics)
    assert test_theta_cn.shape == (len(dataset.test_texts_cn), num_topics)


def test_models(cache_path, num_topics):
    download_dataset("Amazon_Review", cache_path=cache_path)
    download_dataset('dict', cache_path=cache_path)

    dataset = CrosslingualDataset(f"{cache_path}/Amazon_Review", lang1='en', lang2='cn', dict_path=f'{cache_path}/dict/ch_en_dict.dat', as_tensor=True)

    model = topmost.NMTM(
        num_topics=num_topics,
        Map_en2cn=dataset.Map_en2cn,
        Map_cn2en=dataset.Map_cn2en,
        vocab_size_en=dataset.vocab_size_en,
        vocab_size_cn=dataset.vocab_size_cn,
    )

    crosslingual_model_test(model, dataset, num_topics)

    model = topmost.InfoCTM(
        num_topics=num_topics,
        trans_e2c=dataset.trans_matrix_en,
        pretrain_word_embeddings_en=dataset.pretrained_WE_en,
        pretrain_word_embeddings_cn=dataset.pretrained_WE_cn,
        vocab_size_en=dataset.vocab_size_en,
        vocab_size_cn=dataset.vocab_size_cn,
        weight_MI=50
    )

    crosslingual_model_test(model, dataset, num_topics)



import torch
from torch import nn


class ECR(nn.Module):
    '''
        Effective Neural Topic Modeling with Embedding Clustering Regularization. ICML 2023

        Xiaobao Wu, Xinshuai Dong, Thong Thanh Nguyen, Anh Tuan Luu.
    '''
    def __init__(self, weight_loss_ECR, sinkhorn_alpha, OT_max_iter=5000, stopThr=.5e-2):
        super().__init__()

        self.sinkhorn_alpha = sinkhorn_alpha
        self.OT_max_iter = OT_max_iter
        self.weight_loss_ECR = weight_loss_ECR
        self.stopThr = stopThr
        self.epsilon = 1e-16

    def forward(self, M):
        # M: KxV
        # a: Kx1
        # b: Vx1
        device = M.device

        # Sinkhorn's algorithm
        a = (torch.ones(M.shape[0]) / M.shape[0]).unsqueeze(1).to(device)
        b = (torch.ones(M.shape[1]) / M.shape[1]).unsqueeze(1).to(device)

        u = (torch.ones_like(a) / a.size()[0]).to(device) # Kx1

        K = torch.exp(-M * self.sinkhorn_alpha)
        err = 1
        cpt = 0
        while err > self.stopThr and cpt < self.OT_max_iter:
            v = torch.div(b, torch.matmul(K.t(), u) + self.epsilon)
            u = torch.div(a, torch.matmul(K, v) + self.epsilon)
            cpt += 1
            if cpt % 50 == 1:
                bb = torch.mul(v, torch.matmul(K.t(), u))
                err = torch.norm(torch.sum(torch.abs(bb - b), dim=0), p=float('inf'))

        transp = u * (K * v.T)

        loss_ECR = torch.sum(transp * M)
        loss_ECR *= self.weight_loss_ECR

        return loss_ECR



import pytest
import numpy as np

import sys
sys.path.append('../')

from topmost import download_dataset, BasicDataset


@pytest.fixture
def cache_path():
    return './pytest_cache/'


def dataset_test(cache_path, name, data_size, vocab_size, num_labels=None):
    print(name)
    print(num_labels)
    download_dataset(name, cache_path=cache_path)
    dataset = BasicDataset(f"{cache_path}/{name}", read_labels=(num_labels is not None))

    assert (len(dataset.train_texts) + len(dataset.test_texts)) == data_size
    assert dataset.vocab_size == vocab_size

    if num_labels:
        assert len(dataset.train_labels) == len(dataset.train_texts)
        assert len(dataset.test_labels) == len(dataset.test_texts)
        assert len(np.unique(dataset.train_labels)) == num_labels
        assert len(np.unique(dataset.test_labels)) == num_labels


def test_datasets(cache_path):
    dataset_info = [{
            "name": "20NG",
            "num_labels": 20,
            'data_size': 18846,
            'vocab_size': 5000
        },{
            "name": "IMDB",
            "num_labels": 2,
            'data_size': 50000,
            'vocab_size': 5000
        },{
            "name": "NeurIPS",
            'data_size': 7237,
            'vocab_size': 10000
        },{
            "name": "ACL",
            'data_size': 10560,
            'vocab_size': 10000
        },{
            "name": "NYT",
            "num_labels": 12,
            'data_size': 9172,
            'vocab_size': 10000
        }
    ]

    for item in dataset_info:
        print(item)
        dataset_test(cache_path, **item)



import torch
from torch import nn
import torch.nn.functional as F
from .TPD import TPD
from .CDDecoder import CDDecoder
from topmost.models.Encoder import MLPEncoder
from . import utils



class TraCo(nn.Module):
    '''
        On the Affinity, Rationality, and Diversity of Hierarchical Topic Modeling. AAAI 2024

        Xiaobao Wu, Fengjun Pan, Thong Nguyen, Yichao Feng, Chaoqun Liu, Cong-Duy Nguyen, Anh Tuan Luu.
    '''

    def __init__(self, vocab_size, num_topics_list=[10, 50, 200], en_units=300, dropout=0., embed_size=200, bias_topk=20, bias_p=5.0, beta_temp=0.1, weight_loss_TPD=20.0, sinkhorn_alpha=20.0, sinkhorn_max_iter=1000):
        super().__init__()

        self.num_topics_list = num_topics_list
        self.weight_loss_TPD = weight_loss_TPD
        self.beta_temp = beta_temp
        self.num_layers = len(num_topics_list)

        self.bottom_word_embeddings  = nn.init.trunc_normal_(torch.empty(vocab_size, embed_size), std=0.1)
        self.bottom_word_embeddings = nn.Parameter(F.normalize(self.bottom_word_embeddings))

        self.topic_embeddings_list = nn.ParameterList([])
        for num_topic in num_topics_list:
            topic_embeddings = torch.empty((num_topic, self.bottom_word_embeddings.shape[1]))
            nn.init.trunc_normal_(topic_embeddings, std=0.1)
            topic_embeddings = nn.Parameter(F.normalize(topic_embeddings))
            self.topic_embeddings_list.append(topic_embeddings)

        self.TPD = TPD(sinkhorn_alpha, sinkhorn_max_iter)
        self.CDDecoder = CDDecoder(self.num_layers, vocab_size, bias_p, bias_topk)
        self.encoder = MLPEncoder(vocab_size, num_topics_list[-1], en_units, dropout)

        _, self.transp_list = self.TPD(self.topic_embeddings_list)

    def get_beta(self):
        beta_list = list()
        for layer_id, num_topic in enumerate(self.num_topics_list):
            topic_embeddings = self.topic_embeddings_list[layer_id]
            dist = utils.pairwise_euclidean_distance(topic_embeddings, self.bottom_word_embeddings)
            beta = F.softmax(-dist / self.beta_temp, dim=0)
            beta_list.append(beta)

        return beta_list

    def get_phi_list(self):
        transp_list = self.transp_list
        # multiply each transp with the topic size in the next layer.
        phi_list = [item * item.shape[1] for item in transp_list]
        return phi_list

    def get_theta(self, input_bow):
        theta_list = list()
        bottom_theta, mu, logvar = self.encoder(input_bow)
        phi_list = self.get_phi_list()
        # from bottom to top
        for layer_id in range(self.num_layers)[::-1]:
            if layer_id == self.num_layers - 1:
                theta_list.append(bottom_theta)
            else:
                last_theta = theta_list[-1]
                theta = torch.matmul(last_theta, phi_list[layer_id].T)
                theta_list.append(theta)

        theta_list = theta_list[::-1]

        if self.training:
            return theta_list, mu, logvar
        else:
            return theta_list

    def forward(self, input_bow):
        loss = 0.

        loss_TPD, self.transp_list = self.TPD(self.topic_embeddings_list, self.weight_loss_TPD)
        loss += loss_TPD

        theta_list, mu, logvar = self.get_theta(input_bow)
        loss_KL = self.compute_loss_KL(mu, logvar)
        loss_KL = loss_KL.mean()
        loss += loss_KL

        beta_list = self.get_beta()

        recon_loss = self.CDDecoder(input_bow, theta_list, beta_list)
        loss += recon_loss

        rst_dict = {
            'loss': loss,
        }

        return rst_dict

    def compute_loss_KL(self, mu, logvar, mu_prior=None):
        mu_dim = mu.shape[1]
        a = torch.ones((1, mu_dim), device=mu.device)
        if mu_prior is None:
            mu_prior = (torch.log(a).T - torch.mean(torch.log(a), 1)).T
        var_prior = (((1.0 / a) * (1 - (2.0 / mu_dim))).T + (1.0 / (mu_dim * mu_dim)) * torch.sum(1.0 / a, 1)).T

        var = logvar.exp()
        var_division = var / var_prior
        diff = mu - mu_prior
        diff_term = diff * diff / var_prior
        logvar_division = var_prior.log() - logvar
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(axis=1) - mu_dim)
        return KLD



import torch


def get_topk_tensor(matrix, topk, return_mask=False):
    topk_values, topk_idx = torch.topk(matrix, topk)
    topk_tensor = torch.zeros_like(matrix)
    if return_mask:
        topk_tensor.scatter_(1, topk_idx, 1)
    else:
        topk_tensor.scatter_(1, topk_idx, topk_values)

    return topk_tensor


def pairwise_euclidean_distance(x, y):
    cost = torch.sum(x ** 2, axis=-1, keepdim=True) + torch.sum(y ** 2, dim=-1) - 2 * torch.matmul(x, y.t())
    return cost



import pytest

import sys
sys.path.append('../')

from topmost import download_dataset, BasicDataset, DynamicDataset
from topmost import BasicTrainer, DynamicTrainer
from topmost import ETM, DETM
from topmost import eva


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 10


def test_basic_eva(cache_path, num_topics):
    download_dataset("20NG", cache_path=cache_path)
    dataset = BasicDataset(f"{cache_path}/20NG", as_tensor=True)

    model = ETM(num_topics=num_topics, vocab_size=dataset.vocab_size)
    trainer = BasicTrainer(model, dataset, verbose=True, epochs=1)

    top_words = trainer.get_top_words()
    TD = eva._diversity(top_words)
    TC = eva._coherence(dataset.train_texts, dataset.vocab, top_words)
    print("TD: ", TD)
    print("TC: ", TC)


def test_dynamic_evaluations(cache_path, num_topics):
    download_dataset("NYT", cache_path=cache_path)
    dataset = DynamicDataset(f"{cache_path}/NYT", as_tensor=True)

    model = DETM(num_times=dataset.num_times, train_size=dataset.train_size, num_topics=num_topics, vocab_size=dataset.vocab_size, train_time_wordfreq=dataset.train_time_wordfreq)
    trainer = DynamicTrainer(model, dataset, verbose=True, epochs=1)

    top_words = trainer.get_top_words()
    TD = eva.dynamic_diversity(top_words, dataset.train_bow.numpy(), dataset.train_times.numpy(), dataset.vocab)
    TC = eva.dynamic_coherence(dataset.train_texts, dataset.train_times, dataset.vocab, top_words)

    print("TD: ", TD)
    print("TC: ", TC)




import torch
from torch import nn
import torch.nn.functional as F


class DETM(nn.Module):
    """
        The Dynamic Embedded Topic Model. 2019

        Adji B. Dieng, Francisco J. R. Ruiz, David M. Blei
    """
    def __init__(self, vocab_size, num_times, train_size, train_time_wordfreq, num_topics=50, train_WE=True, pretrained_WE=None, en_units=800, eta_hidden_size=200, rho_size=300, enc_drop=0.0, eta_nlayers=3, eta_dropout=0.0, delta=0.005, theta_act='relu', device='cpu'):
        super().__init__()

        ## define hyperparameters
        self.num_topics = num_topics
        self.num_times = num_times
        self.vocab_size = vocab_size
        self.eta_hidden_size = eta_hidden_size
        self.rho_size = rho_size
        self.enc_drop = enc_drop
        self.eta_nlayers = eta_nlayers
        self.t_drop = nn.Dropout(enc_drop)
        self.eta_dropout = eta_dropout
        self.delta = delta
        self.train_WE = train_WE
        self.train_size = train_size
        self.rnn_inp = train_time_wordfreq
        self.device = device

        self.theta_act = self.get_activation(theta_act)

        ## define the word embedding matrix \rho
        if self.train_WE:
            self.rho = nn.Linear(self.rho_size, self.vocab_size, bias=False)
        else:
            rho = nn.Embedding(pretrained_WE.size())
            rho.weight.data = torch.from_numpy(pretrained_WE)
            self.rho = rho.weight.data.clone().float().to(self.device)

        ## define the variational parameters for the topic embeddings over time (alpha) ... alpha is K x T x L
        self.mu_q_alpha = nn.Parameter(torch.randn(self.num_topics, self.num_times, self.rho_size))
        self.logsigma_q_alpha = nn.Parameter(torch.randn(self.num_topics, self.num_times, self.rho_size))

        ## define variational distribution for \theta_{1:D} via amortizartion... theta is K x D
        self.q_theta = nn.Sequential(
            nn.Linear(self.vocab_size + self.num_topics, en_units), 
            self.theta_act,
            nn.Linear(en_units, en_units),
            self.theta_act,
        )
        self.mu_q_theta = nn.Linear(en_units, self.num_topics, bias=True)
        self.logsigma_q_theta = nn.Linear(en_units, self.num_topics, bias=True)

        ## define variational distribution for \eta via amortizartion... eta is K x T
        self.q_eta_map = nn.Linear(self.vocab_size, self.eta_hidden_size)
        self.q_eta = nn.LSTM(self.eta_hidden_size, self.eta_hidden_size, self.eta_nlayers, dropout=self.eta_dropout)
        self.mu_q_eta = nn.Linear(self.eta_hidden_size + self.num_topics, self.num_topics, bias=True)
        self.logsigma_q_eta = nn.Linear(self.eta_hidden_size + self.num_topics, self.num_topics, bias=True)

        self.decoder_bn = nn.BatchNorm1d(vocab_size)
        self.decoder_bn.weight.requires_grad = False

    def get_activation(self, act):
        activations = {
            'tanh': nn.Tanh(),
            'relu': nn.ReLU(),
            'softplus': nn.Softplus(),
            'rrelu': nn.RReLU(),
            'leakyrelu': nn.LeakyReLU(),
            'elu': nn.ELU(),
            'selu': nn.SELU(),
            'glu': nn.GLU(),
        }

        if act in activations:
            act = activations[act]
        else:
            print('Defaulting to tanh activations...')
            act = nn.Tanh()
        return act

    def reparameterize(self, mu, logvar):
        """Returns a sample from a Gaussian distribution via reparameterization.
        """
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return eps.mul_(std).add_(mu)
        else:
            return mu

    def get_kl(self, q_mu, q_logsigma, p_mu=None, p_logsigma=None):
        """Returns KL( N(q_mu, q_logsigma) || N(p_mu, p_logsigma) ).
        """
        if p_mu is not None and p_logsigma is not None:
            sigma_q_sq = torch.exp(q_logsigma)
            sigma_p_sq = torch.exp(p_logsigma)
            kl = ( sigma_q_sq + (q_mu - p_mu)**2 ) / ( sigma_p_sq + 1e-6 )
            kl = kl - 1 + p_logsigma - q_logsigma
            kl = 0.5 * torch.sum(kl, dim=-1)
        else:
            kl = -0.5 * torch.sum(1 + q_logsigma - q_mu.pow(2) - q_logsigma.exp(), dim=-1)
        return kl

    def get_alpha(self): ## mean field
        alphas = torch.zeros(self.num_times, self.num_topics, self.rho_size).to(self.device)
        kl_alpha = []

        alphas[0] = self.reparameterize(self.mu_q_alpha[:, 0, :], self.logsigma_q_alpha[:, 0, :])

        # TODO: why logsigma_p_0 is zero?
        p_mu_0 = torch.zeros(self.num_topics, self.rho_size).to(self.device)
        logsigma_p_0 = torch.zeros(self.num_topics, self.rho_size).to(self.device)
        kl_0 = self.get_kl(self.mu_q_alpha[:, 0, :], self.logsigma_q_alpha[:, 0, :], p_mu_0, logsigma_p_0)
        kl_alpha.append(kl_0)
        for t in range(1, self.num_times):
            alphas[t] = self.reparameterize(self.mu_q_alpha[:, t, :], self.logsigma_q_alpha[:, t, :]) 

            p_mu_t = alphas[t - 1]
            logsigma_p_t = torch.log(self.delta * torch.ones(self.num_topics, self.rho_size).to(self.device))
            kl_t = self.get_kl(self.mu_q_alpha[:, t, :], self.logsigma_q_alpha[:, t, :], p_mu_t, logsigma_p_t)
            kl_alpha.append(kl_t)
        kl_alpha = torch.stack(kl_alpha).sum()
        return alphas, kl_alpha.sum()

    def get_eta(self, rnn_inp): ## structured amortized inference
        inp = self.q_eta_map(rnn_inp).unsqueeze(1)
        hidden = self.init_hidden()
        output, _ = self.q_eta(inp, hidden)
        output = output.squeeze()

        etas = torch.zeros(self.num_times, self.num_topics).to(self.device)
        kl_eta = []

        inp_0 = torch.cat([output[0], torch.zeros(self.num_topics,).to(self.device)], dim=0)
        mu_0 = self.mu_q_eta(inp_0)
        logsigma_0 = self.logsigma_q_eta(inp_0)
        etas[0] = self.reparameterize(mu_0, logsigma_0)

        p_mu_0 = torch.zeros(self.num_topics,).to(self.device)
        logsigma_p_0 = torch.zeros(self.num_topics,).to(self.device)
        kl_0 = self.get_kl(mu_0, logsigma_0, p_mu_0, logsigma_p_0)
        kl_eta.append(kl_0)

        for t in range(1, self.num_times):
            inp_t = torch.cat([output[t], etas[t-1]], dim=0)
            mu_t = self.mu_q_eta(inp_t)
            logsigma_t = self.logsigma_q_eta(inp_t)
            etas[t] = self.reparameterize(mu_t, logsigma_t)

            p_mu_t = etas[t-1]
            logsigma_p_t = torch.log(self.delta * torch.ones(self.num_topics,).to(self.device))
            kl_t = self.get_kl(mu_t, logsigma_t, p_mu_t, logsigma_p_t)
            kl_eta.append(kl_t)
        kl_eta = torch.stack(kl_eta).sum()

        return etas, kl_eta

    def get_theta(self, bows, times, eta=None): ## amortized inference
        """Returns the topic proportions.
        """

        normalized_bows = bows / bows.sum(1, keepdims=True)

        if eta is None and self.training is False:
            eta, kl_eta = self.get_eta(self.rnn_inp)

        eta_td = eta[times]
        inp = torch.cat([normalized_bows, eta_td], dim=1)
        q_theta = self.q_theta(inp)
        if self.enc_drop > 0:
            q_theta = self.t_drop(q_theta)
        mu_theta = self.mu_q_theta(q_theta)
        logsigma_theta = self.logsigma_q_theta(q_theta)
        z = self.reparameterize(mu_theta, logsigma_theta)
        theta = F.softmax(z, dim=-1)
        kl_theta = self.get_kl(mu_theta, logsigma_theta, eta_td, torch.zeros(self.num_topics).to(self.device))

        if self.training:
            return theta, kl_theta
        else:
            return theta

    @property
    def word_embeddings(self):
        return self.rho.weight

    @property
    def topic_embeddings(self):
        alpha, _ = self.get_alpha()
        return alpha

    def get_beta(self, alpha=None):
        """Returns the topic matrix \beta of shape T x K x V
        """

        if alpha is None and self.training is False:
            alpha, kl_alpha = self.get_alpha()

        if self.train_WE:
            logit = self.rho(alpha.view(alpha.size(0) * alpha.size(1), self.rho_size))
        else:
            tmp = alpha.view(alpha.size(0) * alpha.size(1), self.rho_size)
            logit = torch.mm(tmp, self.rho.permute(1, 0))
        logit = logit.view(alpha.size(0), alpha.size(1), -1)

        beta = F.softmax(logit, dim=-1)

        return beta

    def get_NLL(self, theta, beta, bows):
        theta = theta.unsqueeze(1)
        loglik = torch.bmm(theta, beta).squeeze(1)
        loglik = torch.log(loglik + 1e-12)
        nll = -loglik * bows
        nll = nll.sum(-1)
        return nll

    def forward(self, bows, times):
        bsz = bows.size(0)
        coeff = self.train_size / bsz
        eta, kl_eta = self.get_eta(self.rnn_inp)
        theta, kl_theta = self.get_theta(bows, times, eta)
        kl_theta = kl_theta.sum() * coeff

        alpha, kl_alpha = self.get_alpha()
        beta = self.get_beta(alpha)

        beta = beta[times]
        # beta = beta[times.type('torch.LongTensor')]
        nll = self.get_NLL(theta, beta, bows)
        nll = nll.sum() * coeff

        loss = nll + kl_eta + kl_theta

        rst_dict = {
            'loss': loss,
            'nll': nll,
            'kl_eta': kl_eta,
            'kl_theta': kl_theta
        }

        loss += kl_alpha
        rst_dict['kl_alpha'] = kl_alpha

        return rst_dict

    def init_hidden(self):
        """Initializes the first hidden state of the RNN used as inference network for \\eta.
        """
        weight = next(self.parameters())
        nlayers = self.eta_nlayers
        nhid = self.eta_hidden_size
        return (weight.new_zeros(nlayers, 1, nhid), weight.new_zeros(nlayers, 1, nhid))



import numpy as np
from collections import defaultdict
from sklearn import metrics


def purity_score(y_true, y_pred):
    # compute contingency matrix (also called confusion matrix)
    contingency_matrix = metrics.cluster.contingency_matrix(y_true, y_pred)
    return np.sum(np.amax(contingency_matrix, axis=0)) / np.sum(contingency_matrix)


def clustering_metrics(labels, preds):
    metrics_func = [
        {
            'name': 'Purity',
            'method': purity_score
        },
        {
            'name': 'NMI',
            'method': metrics.cluster.normalized_mutual_info_score
        },
    ]

    results = dict()
    for func in metrics_func:
        results[func['name']] = func['method'](labels, preds)

    return results


def _clustering(theta, labels):
    preds = np.argmax(theta, axis=1)
    return clustering_metrics(labels, preds)


def hierarchical_clustering(test_theta, test_labels):
    num_layer = len(test_theta)
    results = defaultdict(list)

    for layer in range(num_layer):
        layer_results = _clustering(test_theta[layer], test_labels)

        for key in layer_results:
            results[key].append(layer_results[key])

    for key in results:
        results[key] = np.mean(results[key])

    return results



from .basic.ProdLDA import ProdLDA
from .basic.CombinedTM import CombinedTM
from .basic.DecTM import DecTM
from .basic.ETM import ETM
from .basic.NSTM.NSTM import NSTM
from .basic.TSCTM.TSCTM import TSCTM
from .basic.ECRTM.ECRTM import ECRTM

from .crosslingual.NMTM import NMTM
from .crosslingual.InfoCTM.InfoCTM import InfoCTM

from .dynamic.DETM import DETM
from .dynamic.CFDTM.CFDTM import CFDTM

from .hierarchical.SawETM.SawETM import SawETM
from .hierarchical.HyperMiner.HyperMiner import HyperMiner
from .hierarchical.TraCo.TraCo import TraCo



import os
import zipfile
from torchvision.datasets.utils import download_url
from topmost.utils.logger import Logger


logger = Logger("WARNING")


def download_dataset(dataset_name, cache_path="~/.topmost"):
    cache_path = os.path.expanduser(cache_path)
    raw_filename = f'{dataset_name}.zip'

    if dataset_name in ['Wikitext-103']:
        # download from Git LFS.
        zipped_dataset_url = f"https://media.githubusercontent.com/media/BobXWu/TopMost/main/data/{raw_filename}"
    else:
        zipped_dataset_url = f"https://raw.githubusercontent.com/BobXWu/TopMost/master/data/{raw_filename}"

    logger.info(zipped_dataset_url)

    download_url(zipped_dataset_url, root=cache_path, filename=raw_filename, md5=None)

    path = f'{cache_path}/{raw_filename}'
    with zipfile.ZipFile(path, 'r') as zip_ref:
        zip_ref.extractall(cache_path)

    os.remove(path)


if __name__ == '__main__':
    download_dataset('20NG')



from gensim.corpora import Dictionary
from gensim.models import CoherenceModel
import numpy as np
from tqdm import tqdm
from ..data.file_utils import split_text_word


def _coherence(reference_corpus, vocab, top_words, cv_type='c_v'):
    split_top_words = split_text_word(top_words)
    num_top_words = len(split_top_words[0])
    for item in split_top_words:
        assert num_top_words == len(item)

    split_reference_corpus = split_text_word(reference_corpus)
    dictionary = Dictionary(split_text_word(vocab))

    cm = CoherenceModel(texts=split_reference_corpus, dictionary=dictionary, topics=split_top_words, topn=num_top_words, coherence=cv_type)
    cv_per_topic = cm.get_coherence_per_topic()
    score = np.mean(cv_per_topic)

    return score


def dynamic_coherence(train_texts, train_times, vocab, top_words_list, cv_type='c_v', verbose=False):
    cv_score_list = list()

    for time, top_words in tqdm(enumerate(top_words_list)):
        # use the texts of each time slice as the reference corpus.
        idx = np.where(train_times == time)[0]
        reference_corpus = [train_texts[i] for i in idx]

        # use the topics at a time slice
        cv_score = _coherence(reference_corpus, vocab, top_words, cv_type)
        cv_score_list.append(cv_score)

    if verbose:
        print(f"dynamic TC list: {cv_score_list}")

    return np.mean(cv_score_list)
